# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Extension of the simple Plot class for displaying segment objects
"""

import operator
import re

import numpy

from matplotlib.artist import allow_rasterization
from matplotlib.ticker import (Formatter, MultipleLocator, NullLocator)
from matplotlib.projections import register_projection
from matplotlib.collections import PatchCollection
from matplotlib.patches import Rectangle
try:
    from mpl_toolkits.axes_grid1 import make_axes_locatable
except ImportError:
    from mpl_toolkits.axes_grid import make_axes_locatable

from .. import version
from ..segments import *
from .timeseries import (TimeSeriesPlot, TimeSeriesAxes)
from .decorators import auto_refresh
from .utils import rUNDERSCORE

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__version__ = version.version


class SegmentAxes(TimeSeriesAxes):
    """Custom `Axes` for a :class:`~gwpy.plotter.SegmentPlot`.

    This `SegmentAxes` provides custom methods for displaying any of

    - :class:`~gwpy.segments.DataQualityFlag`
    - :class:`~gwpy.segments.Segment`
    - :class:`~gwpy.segments.SegmentList`
    - :class:`~gwpy.segments.SegmentListDict`

    Parameters
    ----------
    insetlabels : `bool`, default: `False`
        display segment labels inside the axes. Prevents very long segment
        names from getting squeezed off the end of a standard figure

    See also
    --------
    gwpy.plotter.TimeSeriesAxes
        for documentation of other args and kwargs
    """
    name = 'segments'

    def __init__(self, *args, **kwargs):
        # set labelling format
        kwargs.setdefault('insetlabels', False)

        # make axes
        super(SegmentAxes, self).__init__(*args, **kwargs)

        # set y-axis labels
        self.yaxis.set_major_locator(MultipleLocator())
        formatter = SegmentFormatter()
        self.yaxis.set_major_formatter(formatter)

    # -------------------------------------------------------------------------
    # plotting methods

    def plot(self, *args, **kwargs):
        """Plot data onto these axes

        Parameters
        ----------
        args
            a single instance of

                - :class:`~gwpy.segments.DataQualityFlag`
                - :class:`~gwpy.segments.Segment`
                - :class:`~gwpy.segments.SegmentList`
                - :class:`~gwpy.segments.SegmentListDict`

        kwargs
            keyword arguments applicable to :meth:`~matplotib.axes.Axes.plot`

        Returns
        -------
        Line2D
            the :class:`~matplotlib.lines.Line2D` for this line layer

        See Also
        --------
        :meth:`matplotlib.axes.Axes.plot`
            for a full description of acceptable ``*args` and ``**kwargs``
        """
        out = []
        args = list(args)
        while len(args):
            if isinstance(args[0], DataQualityDict):
                out.append(self.plot_dqdict(args.pop(0), **kwargs))
                continue
            elif isinstance(args[0], DataQualityFlag):
                out.append(self.plot_dqflag(args[0], **kwargs))
                args.pop(0)
                continue
            elif isinstance(args[0], SegmentListDict):
                out.extend(self.plot_segmentlistdict(args[0], **kwargs))
                args.pop(0)
                continue
            elif isinstance(args[0], SegmentList):
                out.append(self.plot_segmentlist(args[0], **kwargs))
                args.pop(0)
                continue
            elif isinstance(args[0], Segment):
                raise ValueError("Input must be DataQualityFlag, "
                                 "SegmentListDict, or SegmentList")
            break
        if len(args):
            out.append(super(SegmentAxes, self).plot(*args, **kwargs))
        self.autoscale(axis='y')
        return out

    @auto_refresh
    def plot_dqdict(self, flags, label='key', known='x', **kwargs):
        """Plot a :class:`~gwpy.segments.DataQualityDict` onto these axes

        Parameters
        ----------
        flags : :class:`~gwpy.segments.DataQualityDict`
            data-quality dict to display
        label : `str`, optional
            labelling system to use, or fixed label for all `DataQualityFlags`.
            Special values include

            - ``'key'``: use the key of the `DataQualityDict`,
            - ``'name'``: use the :attr:`~DataQualityFlag.name` of the
              `DataQualityFlag`

            If anything else, that fixed label will be used for all lines.
        known : `str`, `dict`, `None`, default: '/'
            display `known` segments with the given hatching, or give a
            dict of keyword arguments to pass to
            :meth:`~SegmentAxes.plot_segmentlist`, or `None` to hide.
        **kwargs
            any other keyword arguments acceptable for
            :class:`~matplotlib.patches.Rectangle`

        Returns
        -------
        collection : :class:`~matplotlib.patches.PatchCollection`
            list of :class:`~matplotlib.patches.Rectangle` patches
        """
        out = []
        for lab, flag in flags.iteritems():
            if label.lower() == 'name':
                lab = flag.name
            elif label.lower() != 'key':
                lab = label
            out.append(self.plot(flag, label=rUNDERSCORE.sub(r'\_', lab),
                       **kwargs))
        return out

    @auto_refresh
    def plot_dqflag(self, flag, y=None, known='red', facecolor=(0.2, 0.8, 0.2),
                    **kwargs):
        """Plot a :class:`~gwpy.segments.DataQualityFlag`
        onto these axes

        Parameters
        ----------
        flag : :class:`~gwpy.segments.DataQualityFlag`
            data-quality flag to display
        y : `float`, optional
            y-axis value for new segments
        height : `float`, optional, default: 0.8
            height for each segment block
        known : `str`, `dict`, `None`, default: '/'
            display `known` segments with the given hatching, or give a
            dict of keyword arguments to pass to
            :meth:`~SegmentAxes.plot_segmentlist`, or `None` to hide.
        **kwargs
            any other keyword arguments acceptable for
            :class:`~matplotlib.patches.Rectangle`

        Returns
        -------
        collection : :class:`~matplotlib.patches.PatchCollection`
            list of :class:`~matplotlib.patches.Rectangle` patches
        """
        # get y axis position
        if y is None:
            y = self.get_next_y()
        # get flag name
        name = kwargs.pop('label', flag.texname)

        # get epoch
        try:
            if not self.epoch:
                self.set_epoch(flag.known[0][0])
            else:
                self.set_epoch(min(self.epoch, flag.known[0][0]))
        except IndexError:
            pass
        # make known collection
        if known == 'x' and 'valid' in kwargs:
            known = kwargs.pop('valid')
        if known is not None:
            if isinstance(known, dict):
                vkwargs = known
            else:
                vkwargs = kwargs.copy()
                vkwargs.pop('label', None)
                if known in ['-', '+', 'x', '\\', '*', 'o', 'O', '.']:
                    vkwargs['fill'] = False
                    vkwargs['hatch'] = known
                else:
                    vkwargs['fill'] = True
                    vkwargs['facecolor'] = known
                    vkwargs['edgecolor'] = 'black'
            vkwargs['collection'] = 'ignore'
            vkwargs['zorder'] = -1000
            self.plot_segmentlist(flag.known, y=y, label=name, **vkwargs)
        # make active collection
        collection = self.plot_segmentlist(flag.active, y=y, label=name,
                                           facecolor=facecolor, **kwargs)
        if (known is not None and len(self.collections) == 2 or
                len(self.collections) == 1):
            if len(flag.known):
                self.set_xlim(*map(float, flag.extent))
            self.autoscale(axis='y')
        return collection

    @auto_refresh
    def plot_segmentlist(self, segmentlist, y=None, collection=True,
                         label=None, rasterized=None, **kwargs):
        """Plot a :class:`~gwpy.segments.SegmentList` onto these axes

        Parameters
        ----------
        segmentlist : :class:`~gwpy.segments.SegmentList`
            list of segments to display
        y : `float`, optional
            y-axis value for new segments
        collection : `bool`, default: `True`
            add all patches as a
            :class:`~matplotlib.collections.PatchCollection`, doesn't seem
            to work for hatched rectangles
        label : `str`, optional
            custom descriptive name to print as y-axis tick label
        **kwargs
            any other keyword arguments acceptable for
            :class:`~matplotlib.patches.Rectangle`

        Returns
        -------
        collection : :class:`~matplotlib.patches.PatchCollection`
            list of :class:`~matplotlib.patches.Rectangle` patches
        """
        if y is None:
            y = self.get_next_y()
        patches = []
        for seg in segmentlist:
            patches.append(self.build_segment(seg, y, **kwargs))
        try:
            if not self.epoch:
                self.set_epoch(segmentlist[0][0])
            else:
                self.set_epoch(min(self.epoch, segmentlist[0][0]))
        except IndexError:
            pass
        if collection:
            coll = PatchCollection(patches, len(patches) != 0)
            coll.set_rasterized(rasterized)
            if label is not None:
                coll.set_label(rUNDERSCORE.sub(r'\_', str(label)))
            if collection == 'ignore':
                coll._ignore = True
            else:
                coll._ignore = False
            coll._ypos = y
            out = self.add_collection(coll)
        else:
            out = []
            for p in patches:
                p.set_label(label)
                p.set_rasterized(rasterized)
                label = ''
                out.append(self.add_patch(p))
        self.autoscale(axis='y')
        return out

    @auto_refresh
    def plot_segmentlistdict(self, segmentlistdict, y=None, dy=1, **kwargs):
        """Plot a :class:`~gwpy.segments.SegmentListDict` onto
        these axes

        Parameters
        ----------
        segmentlistdict : :class:`~gwpy.segments.SegmentListDict`
            (name, :class:`~gwpy.segments.SegmentList`) dict
        y : `float`, optional
            starting y-axis value for new segmentlists
        **kwargs
            any other keyword arguments acceptable for
            :class:`~matplotlib.patches.Rectangle`

        Returns
        -------
        collections : `list`
            list of :class:`~matplotlib.patches.PatchCollection` sets for
            each segmentlist
        """
        if y is None:
            y = self.get_next_y()
        collections = []
        for name, segmentlist in segmentlistdict.iteritems():
            collections.append(self.plot_segmentlist(segmentlist, y=y,
                                                     label=name, **kwargs))
            y += dy
        return collections

    @staticmethod
    def build_segment(segment, y, height=.8, valign='center', **kwargs):
        """Build a :class:`~matplotlib.patches.Rectangle` to display
        a single :class:`~gwpy.segments.Segment`

        Parameters
        ----------
        segment : :class:`~gwpy.segments.Segment`
            [start, stop) GPS segment
        y : `float`
            y-axis position for segment
        height : `float`, optional, default: 1
            height (in y-axis units) for segment
        valign : `str`
            alignment of segment on y-axis value:
            `top`, `center`, or `bottom`
        **kwargs
            any other keyword arguments acceptable for
            :class:`~matplotlib.patches.Rectangle`

        Returns
        -------
        box : `~matplotlib.patches.Rectangle`
            rectangle patch for segment display
        """
        if valign.lower() == 'center':
            y0 = y - height/2.
        elif valign.lower() == 'top':
            y0 = y - height
        else:
            raise ValueError("valign must be one of 'top', 'center', or "
                             "'bottom'")
        width = segment[1] - segment[0]
        return Rectangle((segment[0], y0), width=width, height=height,
                         **kwargs)

    def set_xlim(self, *args, **kwargs):
        out = super(SegmentAxes, self).set_xlim(*args, **kwargs)
        _xlim = self.get_xlim()
        try:
            texts = self.texts
        except AttributeError:
            pass
        else:
            for t in texts:
                if hasattr(t, '_is_segment_label') and t._is_segment_label:
                    t.set_x(_xlim[0] + (_xlim[1] - _xlim[0]) * 0.01)
        return out
    set_xlim.__doc__ = TimeSeriesAxes.set_xlim.__doc__

    def get_next_y(self):
        """Find the next y-axis value at which a segment list can be placed

        This method simply counts the number of independent segmentlists or
        flags that have been plotted onto these axes.
        """
        return len(self.get_collections(ignore=False))

    def get_collections(self, ignore=None):
        """Return the collections matching the given `_ignore` value

        Parameters
        ----------
        ignore : `bool`, or `None`
            value of `_ignore` to match

        Returns
        -------
        collections : `list`
            if `ignore=None`, simply returns all collections, otherwise
            returns those collections matching the `ignore` parameter
        """
        if ignore is None:
            return self.collections
        else:
            return [c for c in self.collections if
                    getattr(c, '_ignore', None) == ignore]

    def set_insetlabels(self, inset=None):
        self._insetlabels = (inset is None and not self._insetlabels) or inset

    def get_insetlabels(self):
        """Move the y-axis tick labels inside the axes
        """
        return self._insetlabels

    insetlabels = property(fget=get_insetlabels, fset=set_insetlabels,
                           doc=get_insetlabels.__doc__)

    @allow_rasterization
    def draw(self, *args, **kwargs):
        # inset the labels if requested
        for tick in self.get_yaxis().get_ticklabels():
            if self._insetlabels:
                # record parameters we are changing
                tick._orig_bbox = tick.get_bbox_patch()
                tick._orig_ha = tick.get_ha()
                tick._orig_pos = tick.get_position()
                # modify tick
                tick.set_horizontalalignment('left')
                tick.set_position((0.01, tick.get_position()[1]))
                tick.set_bbox({'alpha': 0.5, 'facecolor': 'white',
                               'edgecolor': 'none'})
            elif self._insetlabels is False:
                # if label has been moved, reset things
                try:
                    tick.set_bbox(tick._orig_bbox)
                except AttributeError:
                    pass
                else:
                    tick.set_horizontalalignment(tick._orig_ha)
                    tick.set_position(tick._orig_pos)
                    del tick._orig_bbox
                    del tick._orig_ha
                    del tick._orig_pos
        return super(SegmentAxes, self).draw(*args, **kwargs)

    draw.__doc__ = TimeSeriesAxes.draw.__doc__

register_projection(SegmentAxes)


class SegmentPlot(TimeSeriesPlot):
    """`Figure` for displaying a :class:`~gwpy.segments.DataQualityFlag`.

    Parameters
    ----------
    *flags : `DataQualityFlag`
        any number of :class:`~gwpy.segments.DataQualityFlag` to
        display on the plot
    insetlabels : `bool`, default: `False`
        display segment labels inside the axes. Prevents very long segment
        names from getting squeezed off the end of a standard figure
    **kwargs
        other keyword arguments as applicable for the
        :class:`~gwpy.plotter.Plot`
    """
    _DefaultAxesClass = SegmentAxes

    def __init__(self, *flags, **kwargs):
        """Initialise a new SegmentPlot
        """
        # separate kwargs into figure args and plotting args
        figargs = {}
        for key in ['figsize', 'auto_refresh']:
            if key in kwargs:
                figargs[key] = kwargs.pop(key)
        sep = kwargs.pop('sep', False)
        epoch = kwargs.pop('epoch', None)
        inset = kwargs.pop('insetlabels', False)

        # generate figure
        super(SegmentPlot, self).__init__(**figargs)

        # plot data
        if len(flags) == 1 and isinstance(flags[0], DataQualityDict):
            flags = flags[0].keys()
        for flag in flags:
            self.add_dataqualityflag(flag,
                                     projection=self._DefaultAxesClass.name,
                                     newax=sep, **kwargs)

        # set epoch
        if len(flags):
            span = reduce(operator.or_, [f.known for f in flags]).extent()
            if not epoch:
                epoch = span[0]
            for ax in self.axes:
                ax.set_epoch(epoch)
                ax.set_xlim(*map(float, span))
                ax.set_insetlabels(inset)
            for ax in self.axes[:-1]:
                ax.set_xlabel("")

        if sep:
            for ax in self.axes:
                ax.set_ylim(-0.5, 0.5)
                ax.grid(b=False, which='both', axis='y')
        elif len(flags):
            ax.set_ylim(-0.5, len(flags)-0.5)
            ax.grid(b=False, which='both', axis='y')

    def add_dataqualityflag(self, flag, **kwargs):
        super(SegmentPlot, self).add_dataqualityflag(flag, **kwargs)
        if self.epoch is None:
            try:
                self.set_epoch(flag.known[0][0])
            except IndexError:
                pass
    add_dataqualityflag.__doc__ = TimeSeriesPlot.add_dataqualityflag.__doc__

    def add_bitmask(self, mask, ax=None, width=0.2, pad=0.1,
                    visible=True, axes_class=SegmentAxes, topdown=False,
                    **plotargs):
        """Display a state-word bitmask on a new set of Axes.
        """
        # find default axes
        if ax is None:
            ax = self.axes[-1]

        # get new segment axes
        divider = make_axes_locatable(ax)
        max = divider.new_horizontal(size=width, pad=pad,
                                     axes_class=axes_class)
        max.set_xscale('gps')
        max.xaxis.set_major_locator(NullLocator())
        max.xaxis.set_minor_locator(NullLocator())
        max.yaxis.set_minor_locator(NullLocator())
        if visible:
            self.add_axes(max)
        else:
            return

        # format mask as a binary string and work out how many bits to set
        if isinstance(mask, int):
            mask = bin(mask)
        elif isinstance(mask, (unicode, str)) and 'x' in mask:
            mask = bin(int(mask, 16))
        maskint = int(mask, 2)
        if topdown:
            bits = list(range(len(mask.split('b', 1)[1])))[::-1]
        else:
            bits = list(range(len(mask.split('b', 1)[1])))

        # loop over bits
        plotargs.setdefault('facecolor', 'green')
        plotargs.setdefault('edgecolor', 'black')
        s = Segment(0, 1)
        for bit in bits:
            if maskint >> bit & 1:
                sl = SegmentList([s])
            else:
                sl = SegmentList()
            max.plot(sl, **plotargs)
        max.set_title('Bitmask')
        max.set_xlim(0, 1)
        max.set_xticks([])
        max.yaxis.set_ticklabels([])
        max.set_xlabel('')
        max.set_ylim(*ax.get_ylim())

        return max


class SegmentFormatter(Formatter):
    """Custom tick formatter for y-axis flag names
    """
    def __init__(self, flags=dict()):
        self.flags = flags

    def __call__(self, t, pos=None):
        # if segments have been plotted at this y-axis value, continue
        for i, coll in enumerate(self.axis.axes.get_collections(ignore=False)):
            if t == coll._ypos:
                return coll.get_label()
        for patch in self.axis.axes.patches:
            if not patch.get_label():
                continue
            if t in Segment(*patch.get_bbox().intervaly):
                return patch.get_label()
        return ''



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2014)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""This module provides methods to calculate the astrophysical sensitive
distance of an instrumental power-spectral density.
"""

import warnings
from math import pi

from scipy import integrate

from astropy import (units, constants)

from .. import version

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__version__ = version.version


def inspiral_range_psd(psd, snr=8, mass1=1.4, mass2=1.4, horizon=False):
    """Compute the inspiral sensitive distance PSD for the given GW strain PSD

    Parameters
    ----------
    psd : `~gwpy.spectrum.Spectrum`
        the instrumental power-spectral-density data
    snr : `float`, optional
        the signal-to-noise ratio for which to calculate range
    mass1 : `float`, `~astropy.units.Quantity`, optional
        the mass (`float` assumed in solar masses) of the first binary
        component
    mass2 : `float`, `~astropy.units.Quantity`, optional
        the mass (`float` assumed in solar masses) of the second binary
    horizon : `bool`, optional
        if `True`, return the maximal 'horizon' sensitive distance, otherwise
        return the angle-averaged range

    Returns
    -------
    rspec : `~gwpy.spectrum.Spectrum`
        the calculated inspiral sensitivity PSD [Mpc^2 / Hz]
    """
    # compute chirp mass and symmetric mass ratio
    mass1 = units.Quantity(mass1, 'solMass').to('kg')
    mass2 = units.Quantity(mass2, 'solMass').to('kg')
    mtotal = mass1 + mass2
    mchirp = (mass1 * mass2) ** (3/5.) / mtotal ** (1/5.)
    # compute ISCO
    fisco = (constants.c ** 3 / (constants.G * 6**1.5 * pi * mtotal)).to('Hz')
    # calculate integral pre-factor
    prefactor = (
        (1.77**2 * 5 * constants.c ** (1/3.) *
            (mchirp * constants.G / constants.c ** 2) ** (5/3.)) /
        (96 * pi ** (4/3.) * snr ** 2)
    )
    # calculate inspiral range ASD in m
    integrand = (
        prefactor.value * psd.frequencies.value ** (-7/3.) / psd /
        units.Mpc.decompose().scale ** 2)
    integrand.override_unit(units.Unit('Mpc^2 / Hz'))
    # restrict to ISCO
    integrand = integrand[psd.frequencies.value < fisco.value]
    # normalize and return
    if integrand.f0.value == 0.0:
        integrand[0] = 0.0
    if horizon:
        integrand *= 2.26 ** 2
    return integrand


def inspiral_range(psd, snr=8, mass1=1.4, mass2=1.4, fmin=0, fmax=None,
                   horizon=False, unit='Mpc'):
    """Calculate the inspiral sensitive distance for the given PSD

    The method returns the distance (in megaparsecs) to which an compact
    binary inspiral with the given component masses would be detectable
    given the instrumental PSD. The calculation is as defined in:

    https://dcc.ligo.org/LIGO-T030276/public

    Parameters
    ----------
    psd : `~gwpy.spectrum.Spectrum`
        the instrumental power-spectral-density data
    snr : `float`, optional
        the signal-to-noise ratio for which to calculate range
    mass1 : `float`, `~astropy.units.Quantity`, optional
        the mass (`float` assumed in solar masses) of the first binary
        component
    mass2 : `float`, `~astropy.units.Quantity`, optional
        the mass (`float` assumed in solar masses) of the second binary
    fmin : `float`, optional
        the lower frequency cut-off of the integral
    fmax : `float`, optional
        the maximum frequency limit of the integral, if not given or `None`,
        the innermost stable circular orbit (ISCO) frequency is used
    horizon : `bool`, optional
        if `True`, return the maximal 'horizon' sensitive distance, otherwise
        return the angle-averaged range

    Returns
    -------
    range : `~astropy.units.Quantity`
        the calculated inspiral range [Mpc]
    """
    mass1 = units.Quantity(mass1, 'solMass').to('kg')
    mass2 = units.Quantity(mass2, 'solMass').to('kg')
    mtotal = mass1 + mass2

    # compute ISCO
    fisco = (constants.c ** 3 / (constants.G * 6**1.5 * pi * mtotal)).to('Hz')
    if not fmax:
        fmax = fisco
    fmax = units.Quantity(fmax, 'Hz')
    if fmax > fisco:
        warnings.warn("Upper frequency bound greater than %s-%s ISCO "
                      "frequency of %s, using ISCO" % (mass1, mass2, fisco))
        fmax = fisco
    fmin = units.Quantity(fmin, 'Hz')

    # integrate
    f = psd.frequencies.to('Hz')
    condition = (f >= fmin) & (f < fmax)
    integrand = inspiral_range_psd(psd[condition], snr=snr, mass1=mass1,
                                   mass2=mass2, horizon=horizon)
    if fmin.value == 0.0:
        integrand[0] = 0.0
    result = units.Quantity(integrate.trapz(integrand.value,
                                            f.value[condition]),
                            unit=integrand.unit * units.Hertz)

    return (result ** (1/2.)).to(unit)


def burst_range_spectrum(psd, snr=8, energy=1e-2, unit='Mpc'):
    """Calculate the frequency-dependent GW burst range for the given PSD

    Parameters
    ----------
    psd : `~gwpy.spectrum.Spectrum`
        the instrumental power-spectral-density data
    snr : `float`, optional
        the signal-to-noise ratio for which to calculate range
    energy : `float`, optional
        the relative energy output of the GW burst, defaults to 1e-2 for
        a GRB-like burst
    unit : `str`, `~astropy.units.Unit`
        desired unit of the returned `~astropy.units.Quantity`

    Returns
    -------
    rangespec : `~gwpy.spectrum.Spectrum`
        the burst range `Spectrum` [Mpc (default)]
    """
    unit = units.Unit(unit)
    a = (constants.G.value * energy * constants.M_sun.value * 0.4 /
         (pi**2 * constants.c.value))**(1/2.)
    dspec = a / (snr * psd**(1/2.) * psd.frequencies) / constants.pc.value
    conv = units.pc.get_converter(unit)
    rspec = conv(dspec)
    rspec.override_unit(unit)
    if rspec.f0.value == 0.0:
        rspec[0] = 0.0
    return rspec


def burst_range(psd, snr=8, energy=1e-2, fmin=100, fmax=500, unit='Mpc'):
    """Calculate the integrated GRB-like burst range for the given PSD

    Parameters
    ----------
    psd : `~gwpy.spectrum.Spectrum`
        the instrumental power-spectral-density data
    snr : `float`, optional
        the signal-to-noise ratio for which to calculate range
    energy : `float`, optional
        the relative energy output of the GW burst, defaults to 1e-2 for
        a GRB-like burst
    fmin : `float`, optional
        the lower frequency cutoff of the burst range integral
    fmax : `float, optional
        the upper frequency cutoff of the burst range integral
    unit : `str`, `~astropy.units.Unit`
        desired unit of the returned `~astropy.units.Quantity`

    Returns
    -------
    range : `~astropy.units.Quantity`
        the GRB-like-burst sensitive range [Mpc (default)]
    """
    freqs = psd.frequencies.value
    # restrict integral
    if not fmin:
        fmin = freqs.min()
    if not fmax:
        fmax = freqs.max()
    condition = (freqs >= fmin) & (freqs < fmax)
    # calculate integrand and integrate
    integrand = burst_range_spectrum(
        psd[condition], snr=snr, energy=energy) ** 3
    result = integrate.trapz(integrand.value, freqs[condition])
    # normalize and return
    r = units.Quantity(result / (fmax - fmin), unit=integrand.unit) ** (1/3.)
    return r.to(unit)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>

"""Git version generator for GWpy (or any package, for that matter)
"""

from __future__ import absolute_import

import os
import time
from distutils.version import (Version, LooseVersion)

from git import Repo
from jinja2 import Template

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__credits__ = 'Adam Mercer <adam.mercer@ligo.org>'


VERSION_PY_TEMPLATE = Template("""# -*- coding: utf-8 -*-
{% if 'author' in package %}\
# Copyright (C) {{ package['author'] }} ({{ package['year'] }})

# package metadata
__author__ = "{{ package['author'] }} <{{ package['email'] }}>"\
{% endif %}
__version__ = '{{ version.vstring }}'
__date__ = '{{ status.datestr }}'

# package version
version = '{{ version.vstring }}'
major = {{ version.version[0] }}
minor = {{ version.version[1] }}
micro = {{ version.version[2] }}
debug = {{ not version.vstring.replace('.', '').isdigit() }}
release = {{ version.version[0] > 0 and \
             version.vstring.replace('.', '').isdigit() }}

# repository version information
git_hash = '{{ status.commit.hexsha }}'
{% if status.tag %}\
git_tag = '{{ status.tag.name }}'\
{% else %}\
git_tag = None\
{% endif %}
{% if status.branch %}\
git_branch = '{{ status.branch.name }}'\
{% else %}
git_branch = None\
{% endif %}
git_author = "{{ status.author }}"
git_committer = "{{ status.committer }}"
git_is_dirty = {{ status.is_dirty() }}
""")


class GitStatus(object):
    def __init__(self, path=os.curdir):
        self.repo = Repo(path=path)

    def is_dirty(self):
        if isinstance(self.repo.is_dirty, bool):
            return self.repo.is_dirty
        else:
            return self.repo.is_dirty()
    is_dirty.__doc__ = Repo.is_dirty.__doc__

    @property
    def commit(self):
        try:
            return self.branch.commit
        except (TypeError, AttributeError):
            return self.repo.head.commit

    @property
    def branch(self):
        try:
            b = self.repo.active_branch
        except TypeError:
            if len(self.repo.branches) == 1:
                return self.repo.branches[0]
            else:
                return None
        else:
            if isinstance(b, str):
                for branch in self.repo.branches:
                    if branch.name == b:
                        return branch
                raise RuntimeError("Cannot resolve active branch.")
            else:
                return b

    @property
    def tag(self):
        for tag in self.tags:
            if tag.commit == self.commit:
                return tag
        return None

    @property
    def tags(self):
        return sorted(self.repo.tags,
                      key=lambda t: t.commit.committed_date)

    @property
    def revision(self):
        """The number of commits between the HEAD and the last tag.
        """
        try:
            tag = self.tags[-1]
        except IndexError:
            start = ''
        else:
            start = '%s..' % tag.name
        return self.repo.git.rev_list('%sHEAD' % start).count('\n')

    @property
    def version(self):
        if self.tag:
            v = self.tag.name.strip('v')
        else:
            try:
                v = self.tags[-1].name.strip('v')
            except IndexError:
                v = '0.0.0'
        if self.is_dirty():
            v += '.dev'
        elif self.revision:
            v += '.dev%d' % self.revision
        return LooseVersion(v)

    @property
    def date(self):
        t = self.commit.committed_date
        if isinstance(t, time.struct_time):
            return t
        else:
            return time.gmtime(t)

    @property
    def datestr(self):
        return time.strftime('%Y-%m-%d %H:%M:%S +0000', self.date)

    @property
    def author(self):
        return "%s <%s>" % (self.commit.author.name, self.commit.author.email)

    @property
    def committer(self):
        return "%s <%s>" % (self.commit.committer.name,
                            self.commit.committer.email)

    # ------------------------------------------------------------------------
    # Write

    def write(self, fobj, version=None, **package_metadata):
        """Write the contents of this `GitStatus` to a file object.
        """
        if version is None:
            version = self.version
        if not isinstance(version, Version):
            version = LooseVersion(str(version))
        if len(version.version) < 3 or not isinstance(version.version[2], int):
            version.version.insert(2, 'None')
        package_metadata.setdefault('year', time.gmtime().tm_year)
        fobj.write(VERSION_PY_TEMPLATE.render(status=self, version=version,
                                              package=package_metadata))



from urllib2 import urlopen
from numpy import asarray
from gwpy.timeseries import TimeSeries
data = urlopen('http://www.ligo.org/science/GW100916/'
               'L-strain_hp30-968654552-10.txt').read()
ts = TimeSeries(asarray(data.splitlines(), dtype=float),
                epoch=968654552, sample_rate=16384)
plot = ts.plot()
plot.set_ylabel('Gravitational-wave strain amplitude')
plot.set_title('LIGO Livingston Observatory data for GW100916')
plot.show()



#!/usr/bin/env python

# Copyright (C) Duncan Macleod (2014)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Calculating (and plotting) event rate versus time

I would like to study the rate at which event triggers are generated by the
`ExcessPower` gravitational-wave burst detection algorithm, over a small
stretch of data.

The data from which these events were generated are a simulation of Gaussian noise
with the Advanced LIGO design spectrum, and so don't actually contain any real
gravitational waves, but will help tune the algorithm to improve detection of
future, real signals.
"""

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__currentmodule__ = 'gwpy.table.lsctables'

# First, import the `SnglBurstTable`:
from gwpy.table.lsctables import SnglBurstTable

# and read a table of (simulated) events:
events = SnglBurstTable.read('../../gwpy/tests/data/'
                             'H1-LDAS_STRAIN-968654552-10.xml.gz')

# We can calculate the rate of events (in Hertz) using the :meth:`~SnglBurstTable.event_rate` method:
rate = events.event_rate(1, start=968654552, end=968654562)

# The :meth:`~SnglBurstTable.event_rate` method has returned a :class:`~gwpy.timeseries.TimeSeries`, so we can display this using the :meth:`~gwpy.timeseries.TimeSeries.plot` method of that object:
plot = rate.plot()
plot.set_xlim(968654552, 968654562)
plot.set_ylabel('Event rate [Hz]')
plot.set_title('LIGO Hanford Observatory event rate for GW100916')
plot.show()



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""This module defines the Boolean array representing the state of some data

Such states are typically the comparison of a `TimeSeries` against some
threshold, where sub-threshold is good and sup-threshold is bad,
for example.

Single `StateTimeSeries` can be bundled together to form `StateVector`
arrays, representing a bit mask of states that combine to make a detailed
statement of instrumental operation
"""

from math import (ceil, log)
import sys

import numpy

from glue.segmentsUtils import from_bitstream
from astropy.units import Quantity

from .core import (TimeSeriesBase, TimeSeriesBaseDict, TimeSeriesBaseList,
                   ArrayTimeSeries, NDS2_FETCH_TYPE_MASK,
                   as_series_dict_class)
from ..detector import Channel
from ..time import Time
from ..io import reader
from .. import version
from ..utils.docstring import interpolate_docstring

if sys.version_info[0] < 3:
    range = xrange

__version__ = version.version
__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

__all__ = ['StateTimeSeries',
           'StateVector', 'StateVectorDict', 'StateVectorList', 'Bits']


@interpolate_docstring
class StateTimeSeries(TimeSeriesBase):
    """Boolean array representing a good/bad state determination
    of some data.

    Parameters
    ----------
    %(Array1)s

    %(time-axis)s

    %(Array2)s

    Notes
    -----
    The input data array is cast to the `bool` data type upon creation of
    this series.

    .. rubric:: Key methods

       ~StateTimeSeries.to_dqflag

    """

    def __new__(cls, data, times=None, epoch=None, channel=None,
                unit='dimensionless', sample_rate=None, name=None, **kwargs):
        """Generate a new StateTimeSeries
        """
        if isinstance(data, (list, tuple)):
            data = numpy.asarray(data)
        if not isinstance(data, cls):
            data = data.astype(bool)
        return super(StateTimeSeries, cls).__new__(cls, data, name=name,
                                                   epoch=epoch,
                                                   channel=channel,
                                                   sample_rate=sample_rate,
                                                   times=times, **kwargs)

    def to_dqflag(self, name=None, minlen=1, dtype=float, round=False,
                  label=None, description=None):
        """Convert this `StateTimeSeries` into a
        `~gwpy.segments.DataQualityFlag`

        Each contiguous set of `True` values are grouped as a
        `~gwpy.segments.Segment` running from the GPS time the first
        found `True`, to the GPS time of the next `False` (or the end
        of the series)

        Parameters
        ----------
        minlen : `int`, optional, default: 1
            minimum number of consecutive `True` values to identify as a
            `~gwpy.segments.Segment`. This is useful to ignore single
            bit flips, for example.

        dtype : `type`, `callable`, default: `float`
            output segment entry type, can pass either a type for simple
            casting, or a callable function that accepts a float and returns
            another numeric type

        round : `bool`, optional, default: False
            choose to round each `~gwpy.segments.Segment` to its
            inclusive integer boundaries

        Returns
        -------
        dqflag : `~gwpy.segments.DataQualityFlag`
            a segment representation of this `StateTimeSeries`, the span
            defines the `known` segments, while the contiguous `True`
            sets defined each of the `active` segments
        """
        from ..segments import (Segment, SegmentList, DataQualityFlag)
        start = self.x0.value
        dt = self.dx.value
        active = from_bitstream(self.value, start, dt, minlen=int(minlen))
        if dtype is not float:
            active = active.__class__([Segment(dtype(s[0]), dtype(s[1])) for
                                       s in active])
        known = SegmentList([self.span])
        out = DataQualityFlag(name=name or self.name, active=active,
                              known=known, label=label or self.name,
                              description=description)
        if round:
            out = out.round()
        return out.coalesce()

    def to_lal(self, *args, **kwargs):
        """Bogus function inherited from superclass, do not use.
        """
        raise NotImplementedError("The to_lal method, inherited from the "
                                  "TimeSeries, cannot be used with the "
                                  "StateTimeSeries because LAL has no "
                                  "BooleanTimeSeries structure")

    def __getitem__(self, item):
        if isinstance(item, (float, int)):
            return numpy.ndarray.__getitem__(self, item)
        else:
            return super(StateTimeSeries, self).__getitem__(item)


class Bits(list):
    """Definition of the bits in a `StateVector`.

    Parameters
    ----------
    bits : `list`
        list of bit names
    channel : `Channel`, `str`, optional
        data channel associated with this Bits
    epoch : `float`, optional
        defining GPS epoch for this `Bits`
    description : `dict`, optional
        (bit, desc) `dict` of longer descriptions for each bit
    """
    def __init__(self, bits, channel=None, epoch=None, description=None):
        list.__init__(self, [b or None for b in bits])
        if channel is not None:
            self.channel = channel
        if epoch is not None:
            self.epoch = epoch
        self.description = description
        for i, bit in enumerate(bits):
            if bit is None or bit in self.description:
                continue
            elif channel:
                self.description[bit] = '%s bit %d' % (self.channel, i)
            else:
                self.description[bit] = None

    @property
    def epoch(self):
        """Starting GPS time epoch for these `Bits`.

        This attribute is recorded as a `~astropy.time.Time` object in the
        GPS format, allowing native conversion into other formats.

        See :mod:`~astropy.time` for details on the `Time` object.
        """
        try:
            return Time(self._epoch, format='gps')
        except AttributeError:
            return None

    @epoch.setter
    def epoch(self, epoch):
        if isinstance(epoch, Time):
            self._epoch = epoch.gps
        elif isinstance(epoch, Quantity):
            self._epoch = epoch.value
        else:
            self._epoch = float(epoch)

    @property
    def channel(self):
        """Data channel associated with these `Bits`.
        """
        try:
            return self._channel
        except AttributeError:
            return None

    @channel.setter
    def channel(self, ch):
        if isinstance(ch, Channel):
            self._channel = ch
        else:
            self._channel = Channel(ch)

    @property
    def description(self):
        """(key, value) dictionary of long bit descriptions.
        """
        return self._description

    @description.setter
    def description(self, desc):
        if desc is None:
            self._description = {}
        else:
            self._description = desc

    def __repr__(self):
        indent = " " * len('<%s(' % self.__class__.__name__)
        mask = ('\n%s' % indent).join(['%d: %r' % (idx, bit) for
                                       idx, bit in enumerate(self)
                                       if bit])
        return ("<{1}({2},\n{0}channel={3},\n{0}epoch={4})>".format(
                indent, self.__class__.__name__,
                mask, repr(self.channel), repr(self.epoch)))

    def __str__(self):
        indent = " " * len('%s(' % self.__class__.__name__)
        mask = ('\n%s' % indent).join(['%d: %s' % (idx, bit) for
                                       idx, bit in enumerate(self)
                                       if bit])
        return ("{1}({2},\n{0}channel={3},\n{0}epoch={4})".format(
                indent, self.__class__.__name__,
                mask, str(self.channel), str(self.epoch)))

    def __array__(self):
        return numpy.array([b or '' for b in self])


@interpolate_docstring
class StateVector(TimeSeriesBase):
    """Binary array representing good/bad state determinations of some data.

    Each binary bit represents a single boolean condition, with the
    definitions of all the bits stored in the `StateVector.bits`
    attribute.

    Parameters
    ----------
    %(Array1)s

    bits : `Bits`, `list`, optional
        list of bits defining this `StateVector`

    %(time-axis)s

    %(Array2)s

    Notes
    -----
    Key methods:

    .. autosummary::

        ~StateVector.fetch
        ~StateVector.read
        ~StateVector.write
        ~StateVector.to_dqflags
        ~StateVector.plot

    """
    _metadata_slots = TimeSeriesBase._metadata_slots + ['bits']

    def __new__(cls, data, bits=None, times=None, epoch=None, sample_rate=None,
                channel=None, name=None, **kwargs):
        """Generate a new `StateVector`.
        """
        new = super(StateVector, cls).__new__(cls, data, name=name,
                                              epoch=epoch, channel=channel,
                                              sample_rate=sample_rate,
                                              times=times,
                                              **kwargs)
        new.bits = bits
        return new

    # -------------------------------------------
    # StateVector properties

    @property
    def bits(self):
        """The list of bit names for this `StateVector`.

        :type: `Bits`
        """
        try:
            return self._bits
        except AttributeError as e:
            if self.dtype.name.startswith(('uint', 'int')):
                nbits = self.itemsize * 8
                self.bits = Bits(['Bit %d' % b for b in range(nbits)],
                                 channel=self.channel, epoch=self.epoch)
                return self.bits
            elif hasattr(self.channel, 'bits'):
                self.bits = self.channel.bits
                return self.bits
            else:
                return None

    @bits.setter
    def bits(self, mask):
        if mask is None:
            del self.bits
            return
        if not isinstance(mask, Bits):
            mask = Bits(mask, channel=self.channel,
                        epoch=self.epoch)
        self._bits = mask

    @bits.deleter
    def bits(self):
        try:
            del self._bits
        except AttributeError:
            pass

    @property
    def boolean(self):
        """A mapping of this `StateVector` to a 2-D array containing all
        binary bits as booleans, for each time point.
        """
        try:
            return self._boolean
        except AttributeError:
            nbits = len(self.bits)
            boolean = numpy.zeros((self.size, nbits), dtype=bool)
            for i, d in enumerate(self.value):
                boolean[i, :] = [int(d) >> j & 1 for
                                 j in range(nbits)]
            self._boolean = ArrayTimeSeries(boolean, name=self.name,
                                            epoch=self.epoch,
                                            sample_rate=self.sample_rate,
                                            y0=0, dy=1)
            return self.boolean

    def get_bit_series(self, bits=None):
        """Get the `StateTimeSeries` for each bit of this `StateVector`.

        Parameters
        ----------
        bits : `list`, optional
            a list of bit indices or bit names, defaults to
            `~StateVector.bits`

        Returns
        -------
        bitseries : `TimeSeriesDict`
            a `TimeSeriesDict` of `StateTimeSeries`, one for each given
            bit
        """
        if bits is None:
            bits = [b for b in self.bits if b is not None and b is not '']
        bindex = []
        for b in bits:
            try:
                bindex.append((self.bits.index(b), b))
            except IndexError as e:
                e.args = ('Bit %r not found in StateVector' % b)
                raise e
        self._bitseries = StateTimeSeriesDict()
        for i, bit in bindex:
            self._bitseries[bit] = StateTimeSeries(
                self.value >> i & 1, name=bit, epoch=self.x0.value,
                channel=self.channel, sample_rate=self.sample_rate)
        return self._bitseries

    # -------------------------------------------
    # StateVector methods

    # use input/output registry to allow multi-format reading
    read = classmethod(interpolate_docstring(
        reader(doc="""Read data into a `StateVector`

        Parameters
        ----------
        %(timeseries-read1)s

        bits : `list`, optional
            list of bits names for this `StateVector`, give `None` at
            any point in the list to mask that bit

        %(timeseries-read2)s

        Example
        -------
        To read the S6 state vector, with names for all the bits::

            >>> sv = StateVector.read(
                'H-H1_LDAS_C02_L2-968654592-128.gwf', 'H1:IFO-SV_STATE_VECTOR',
                bits=['Science mode', 'Conlog OK', 'Locked',
                      'No injections', 'No Excitations'],
                dtype='uint32')

        then you can convert these to segments

            >>> segments = sv.to_dqflags()

        or to read just the interferometer operations bits::

            >>> sv = StateVector.read(
                'H-H1_LDAS_C02_L2-968654592-128.gwf', 'H1:IFO-SV_STATE_VECTOR',
                bits=['Science mode', None, 'Locked'], dtype='uint32')

        Running `to_dqflags` on this example would only give 2 flags, rather
        than all five.

        Alternatively the `bits` attribute can be reset after reading, but
        before any further operations.

        Notes
        -----""")))

    def to_dqflags(self, bits=None, minlen=1, dtype=float, round=False):
        """Convert this `StateVector` into a `~gwpy.segments.DataQualityDict`

        The `StateTimeSeries` for each bit is converted into a
        `~gwpy.segments.DataQualityFlag` with the bits combined into a dict.

        Parameters
        ----------
        minlen : `int`, optional, default: 1
           minimum number of consecutive `True` values to identify as a
           `Segment`. This is useful to ignore single bit flips,
           for example.

        bits : `list`, optional
            a list of bit indices or bit names to select, defaults to
            `~StateVector.bits`

        Returns
        -------
        DataQualityFlag list : `list`
            a list of :class:`~gwpy.segments.flag.DataQualityFlag`
            reprensentations for each bit in this `StateVector`

        See Also
        --------
        :meth:`StateTimeSeries.to_dqflag`
            for details on the segment representation method for
            `StateVector` bits
        """
        from ..segments import DataQualityDict
        out = DataQualityDict()
        bitseries = self.get_bit_series(bits=bits)
        for bit, sts in bitseries.iteritems():
            out[bit] = sts.to_dqflag(name=bit, minlen=minlen, round=round,
                                     dtype=dtype,
                                     description=self.bits.description[bit])
        return out

    @classmethod
    @interpolate_docstring
    def fetch(cls, channel, start, end, bits=None, host=None, port=None,
              verbose=False, connection=None, type=NDS2_FETCH_TYPE_MASK):
        """Fetch data from NDS into a `StateVector`.

        Parameters
        ----------
        %(timeseries-fetch1)s

        bits : `Bits`, `list`, optional
            definition of bits for this `StateVector`

        %(timeseries-fetch2)s
        """
        new = StateVectorDict.fetch(
            [channel], start, end, host=host, port=port,
            verbose=verbose, connection=connection)[channel]
        if bits:
            new.bits = bits
        return new

    def to_lal(self, *args, **kwargs):
        """Bogus function inherited from superclass, do not use.
        """
        raise NotImplementedError("The to_lal method, inherited from the "
                                  "TimeSeries, cannot be used with the "
                                  "StateTimeSeries because LAL has no "
                                  "BooleanTimeSeries structure")

    def plot(self, format='segments', bits=None, **kwargs):
        """Plot the data for this `StateVector`

        Parameters
        ----------
        format : `str`, optional, default: ``'segments'``
            type of plot to make, either 'segments' to plot the
            SegmentList for each bit, or 'timeseries' to plot the raw
            data for this `StateVector`
        bits : `list`, optional
            a list of bit indices or bit names, defaults to
            `~StateVector.bits`. This argument is ignored if ``format`` is
            not ``'segments'``
        **kwargs
            other keyword arguments to be passed to either
            :class:`~gwpy.plotter.segments.SegmentPlot` or
            :class:`~gwpy.plotter.timeseries.TimeSeriesPlot`, depending
            on ``format``.

        Returns
        -------
        plot : :class:`~gwpy.plotter.segments.SegmentPlot`, or
               :class:`~gwpy.plotter.timeseries.TimeSeriesPlot`
            output plot object, subclass of :class:`~gwpy.plotter.core.Plot`
        """
        if format == 'timeseries':
            return super(StateVector, self).plot(**kwargs)
        elif format == 'segments':
            kwargs.setdefault('facecolor', 'green')
            kwargs.setdefault('edgecolor', 'black')
            kwargs.setdefault('known', {'facecolor': 'red',
                                        'edgecolor': 'black'})
            from ..plotter import SegmentPlot
            return SegmentPlot(*self.to_dqflags(bits=bits).values(), **kwargs)
        raise ValueError("'format' argument must be one of: 'timeseries' or "
                         "'segments'")

    def resample(self, rate):
        """Resample this `StateVector` to a new rate

        Because of the nature of a state-vector, downsampling is done
        by taking the logical 'and' of all original samples in each new
        sampling interval, while upsampling is achieved by repeating
        samples.

        Parameters
        ----------
        rate : `float`
            rate to which to resample this `StateVector`, must be a
            divisor of the original sample rate (when downsampling)
            or a multiple of the original (when upsampling).

        Returns
        -------
        vector : :class:`~gwpy.timeseries.statevector.StateVector`
            resampled version of the input `StateVector`
        """
        rate1 = self.sample_rate.value
        if isinstance(rate, Quantity):
            rate2 = rate.value
        else:
            rate2 = float(rate)
        # upsample
        if (rate2 / rate1).is_integer():
            raise NotImplementedError("StateVector upsampling has not "
                                      "been implemented yet, sorry.")
        # downsample
        elif (rate1 / rate2).is_integer():
            factor = int(rate1 / rate2)
            # reshape incoming data to one column per new sample
            newsize = self.size / factor
            old = self.value.reshape((newsize, self.size // newsize))
            # work out number of bits
            if len(self.bits):
                nbits = len(self.bits)
            else:
                max = self.value.max()
                nbits = max != 0 and int(ceil(log(self.value.max(), 2))) or 1
            bits = range(nbits)
            # construct an iterator over the columns of the old array
            it = numpy.nditer([old, None],
                              flags=['external_loop', 'reduce_ok'],
                              op_axes=[None, [0, -1]],
                              op_flags=[['readonly'],
                                        ['readwrite', 'allocate']])
            dtype = self.dtype
            type_ = self.dtype.type
            # for each new sample, each bit is logical AND of old samples
            # bit is ON,
            for x, y in it:
                y[...] = numpy.sum([type_((x >> bit & 1).all() * (2 ** bit))
                                    for bit in bits], dtype=self.dtype)
            new = StateVector(it.operands[1], dtype=dtype)
            new.__dict__ = self.copy_metadata()
            new.sample_rate = rate2
            return new
        # error for non-integer resampling factors
        elif rate1 < rate2:
            raise ValueError("New sample rate must be multiple of input "
                             "series rate if upsampling a StateVector")
        else:
            raise ValueError("New sample rate must be divisor of input "
                             "series rate if downsampling a StateVector")


@as_series_dict_class(StateTimeSeries)
class StateTimeSeriesDict(TimeSeriesBaseDict):
    EntryClass = StateTimeSeries
    read = classmethod(reader(doc=TimeSeriesBaseDict.read.__doc__))


@as_series_dict_class(StateVector)
class StateVectorDict(TimeSeriesBaseDict):
    EntryClass = StateVector
    read = classmethod(reader(doc="""
        Read data for multiple bit vector channels into a `StateVectorDict`

        Parameters
        ----------
        source : `str`, `~glue.lal.Cache`
            a single file path `str`, or a `~glue.lal.Cache` containing
            a contiguous list of files.

        channels : `~gwpy.detector.channel.ChannelList`, `list`
            a list of channels to read from the source.

        start : `~gwpy.time.LIGOTimeGPS`, `float`, `str` optional
            GPS start time of required data, anything parseable by
            :meth:`~gwpy.time.to_gps` is fine

        end : `~gwpy.time.LIGOTimeGPS`, `float`, `str`, optional
            GPS end time of required data, anything parseable by
            :meth:`~gwpy.time.to_gps` is fine

        bits : `list` of `lists`, `dict`, optional
            the ordered list of interesting bit lists for each channel,
            or a `dict` of (`channel`, `list`) pairs

        format : `str`, optional
            source format identifier. If not given, the format will be
            detected if possible. See below for list of acceptable
            formats.

        nproc : `int`, optional, default: ``1``
            number of parallel processes to use, serial process by
            default.

            .. note::

               Parallel frame reading, via the ``nproc`` keyword argument,
               is only available when giving a :class:`~glue.lal.Cache` of
               frames, or using the ``format='cache'`` keyword argument.

        gap : `str`, optional
            how to handle gaps in the cache, one of

            - 'ignore': do nothing, let the undelying reader method handle it
            - 'warn': do nothing except print a warning to the screen
            - 'raise': raise an exception upon finding a gap (default)
            - 'pad': insert a value to fill the gaps

        pad : `float`, optional
            value with which to fill gaps in the source data, only used if
            gap is not given, or `gap='pad'` is given

        Returns
        -------
        statevectordict : `StateVectorDict`
            a `StateVectorDict` of (`channel`, `StateVector`) pairs. The keys
            are guaranteed to be the ordered list `channels` as given.

        Notes
        -----
        """))


class StateVectorList(TimeSeriesBaseList):
    EntryClass = StateVector



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""`Spectrum` calculation methods using the SciPy module.
"""

import numpy

from astropy import units

from .core import Spectrum
from .registry import register_method
from ..utils import import_method_dependency
from .utils import scale_timeseries_units
from .. import version
__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__version__ = version.version


def welch(timeseries, segmentlength, noverlap=None, **kwargs):
    """Calculate the PSD using the scipy Welch method.
    """
    # get module
    signal = import_method_dependency('scipy.signal')
    # calculate PSD
    f, psd_ = signal.welch(timeseries.value, noverlap=noverlap,
                           fs=timeseries.sample_rate.decompose().value,
                           nperseg=segmentlength, **kwargs)
    # generate Spectrum and return
    unit = scale_timeseries_units(timeseries.unit,
                                  kwargs.get('scaling', 'density'))
    return Spectrum(psd_, unit=unit, frequencies=f, name=timeseries.name,
                    epoch=timeseries.epoch, channel=timeseries.channel)

register_method(welch)


def bartlett(timeseries, segmentlength, **kwargs):
    """Calculate a PSD using the Bartlett average method.
    """
    import_method_dependency('scipy.signal')
    kwargs.pop('noverlap', None) 
    return welch(timeseries, segmentlength, noverlap=0, **kwargs)

register_method(bartlett)


def rayleigh(timeseries, segmentlength, noverlap=0, **kwargs):
    """Calculate a Rayleigh statistic spectrum
    """
    stepsize = segmentlength - noverlap
    if noverlap:
        numsegs = 1 + int((timeseries.size - segmentlength) / float(noverlap))
    else:
        numsegs = int(timeseries.size // segmentlength)
    tmpdata = numpy.ndarray((numsegs, int(segmentlength//2 + 1)))
    for i in range(numsegs):
        ts = timeseries[i*stepsize:i*stepsize+segmentlength]
        tmpdata[i, :] = welch(ts, segmentlength)
    std = tmpdata.std(axis=0)
    mean = tmpdata.mean(axis=0)
    return Spectrum(std/mean, unit='', copy=False, f0=0,
                    df=timeseries.sample_rate.value/segmentlength,
                    channel=timeseries.channel,
                    name='Rayleigh spectrum of %s' % timeseries.name)

register_method(rayleigh)


def csd(timeseries, othertimeseries, segmentlength, noverlap=None, **kwargs):
    """Calculate the CSD using scipy's csd method (which uses Welch's method)
    """
    # get module
    signal = import_method_dependency('scipy.signal')
    # calculate CSD
    f, csd_ = signal.csd(timeseries.value, othertimeseries.value,
                         noverlap=noverlap,
                         fs=timeseries.sample_rate.decompose().value,
                         nperseg=segmentlength, **kwargs)
    # generate Spectrum and return
    unit = scale_timeseries_units(timeseries.unit,
                                  kwargs.get('scaling', 'density'))
    return Spectrum(csd_, unit=unit, frequencies=f,
                    name=str(timeseries.name)+'---'+str(othertimeseries.name),
                    epoch=timeseries.epoch, channel=timeseries.channel)

register_method(csd)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) Joseph Areeda (2015)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.
#

""" Coherence plots
"""
from cliproduct import CliProduct


class Spectrogram(CliProduct):
    """Derived class to calculate Spectrograms"""

    def get_action(self):
        """Return the string used as "action" on command line."""
        return 'spectrogram'

    def init_cli(self, parser):
        """Set up the argument list for this product"""
        self.arg_chan1(parser)
        self.arg_freq(parser)
        self.arg_ax_linx(parser)
        self.arg_ax_ylf(parser)
        self.arg_ax_intlog(parser)
        self.arg_imag(parser)
        self.arg_plot(parser)
        return

    def get_ylabel(self, args):
        """Default text for y-axis label"""
        return 'Frequency (Hz)'

    def get_color_label(self):
        return self.scaleText

    def get_max_datasets(self):
        """Spectrogram only handles 1 at a time"""
        return 1

    def is_image(self):
        """This plot is image type"""
        return True

    def freq_is_y(self):
        """This plot puts frequency on the y-axis of the image"""
        return True

    def get_title(self):
        """Start of default super title, first channel is appended to it"""
        return 'Spectrogram: '

    def gen_plot(self, arg_list):
        """Generate the plot from time series and arguments"""
        self.is_freq_plot = True

        from numpy import percentile

        secpfft = 1
        if arg_list.secpfft:
            secpfft = float(arg_list.secpfft)
        ovlp_frac = 0.5
        if arg_list.overlap:
            ovlp_frac = float(arg_list.overlap)
        self.secpfft = secpfft
        self.overlap = ovlp_frac

        ovlp_sec = secpfft*ovlp_frac
        nfft = self.dur/(secpfft - ovlp_sec)
        fft_per_stride = int(nfft/(self.width * 0.8))
        stride_sec = fft_per_stride * (secpfft - ovlp_sec) + secpfft - 1
        stride_sec = max(2*secpfft, stride_sec)
        fs = self.timeseries[0].sample_rate.value

        # based on the number of FFT calculation choose between
        # high time resolution and high SNR
        snr_nfft = self.dur / stride_sec
        if snr_nfft > 512:
            specgram = self.timeseries[0].spectrogram(stride_sec,
                                                      fftlength=secpfft,
                                                      overlap=ovlp_sec)
            self.log(3, ('Spectrogram calc, stride: %.2fs, fftlength: %.2f, '
                         'overlap: %.2f, #fft: %d' %
                         (stride_sec, secpfft, ovlp_sec, snr_nfft)))
        else:
            specgram = self.timeseries[0].spectrogram2(fftlength=secpfft,
                                                       overlap=ovlp_sec)
            self.log(3, ('HR-Spectrogram calc, stride: %.2fs, fftlength: %.2f,'
                         ' overlap: %.2f, #fft: %d' %
                         (stride_sec, secpfft, ovlp_sec, snr_nfft)))
        specgram = specgram ** (1/2.)   # ASD

        norm = False
        if arg_list.norm:
            specgram = specgram.ratio('median')
            norm = True

        # set default frequency limits
        self.fmax = fs / 2.
        self.fmin = 1 / secpfft

        # default time axis
        self.xmin = self.timeseries[0].times.value.min()
        self.xmax = self.timeseries[0].times.value.max()

        # set intensity (color) limits
        if arg_list.imin:
            lo = float(arg_list.imin)
        else:
            lo = .01
        if norm or arg_list.nopct:
            imin = lo
        else:
            imin = percentile(specgram, lo*100)

        if arg_list.imax:
            up = float(arg_list.imax)
        else:
            up = 100
        if arg_list.nopct:
            imax = up
        else:
            imax = percentile(specgram, up)

        if norm:
            self.plot = specgram.plot(norm='log', vmin=imin, vmax=imax)
            self.scaleText = 'Normalized to median'
        elif arg_list.lincolors:
            self.plot = specgram.plot(vmin=imin, vmax=imax)
            self.scaleText = r'ASD $\left( \frac{\mathrm{Counts}}' \
                             r'{\sqrt{\mathrm{Hz}}}\right)$'
        else:
            self.plot = specgram.plot(norm='log', vmin=imin, vmax=imax)
            self.scaleText = r'$ASD \left(\frac{\mathrm{Counts}}' \
                             r'{\sqrt{\mathrm{Hz}}}\right)$'
        # pass the image limits back to the annotater
        self.imin = imin
        self.imax = imax
        return



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""The `Series` is a one-dimensional array with metadata
"""

from warnings import warn
from math import floor

import numpy

from astropy.units import (Unit, Quantity, dimensionless_unscaled)

from .array import Array
from .. import version
from ..utils.docstring import interpolate_docstring

__version__ = version.version
__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


interpolate_docstring.update(
    ArrayXaxis=(
        """x0 : `float`, `~astropy.units.Quantity`, optional, default: `0`
        the starting value for the x-axis of this array

    dx : `float`, `~astropy.units.Quantity, optional, default: `1`
        the step size for the x-axis of this array

    xindex : `array-like`
        the complete array of x-axis values for this array. This argument
        takes precedence over `x0` and `dx` so should be
        given in place of these if relevant, not alongside"""),
)


@interpolate_docstring
class Series(Array):
    """A one-dimensional data series

    Parameters
    ----------
    %(Array1)s

    %(ArrayXaxis)s

    %(Array2)s

    Returns
    -------
    series : `Series`
        a new `Series`
    """
    _metadata_slots = Array._metadata_slots + ['x0', 'dx']
    _default_xunit = Unit('')
    _ndim = 1

    def __new__(cls, value, unit=None, xindex=None, x0=0, dx=1, **kwargs):
        shape = numpy.shape(value)
        if len(shape) != cls._ndim:
            raise ValueError("Cannot generate %s with %d-dimensional data"
                             % (cls.__name__, len(shape)))
        new = super(Series, cls).__new__(cls, value, unit=unit, **kwargs)
        if isinstance(x0, Quantity):
            xunit = x0.unit
        elif isinstance(dx, Quantity):
            xunit = dx.unit
        else:
            xunit = cls._default_xunit
        new.x0 = Quantity(x0, xunit)
        new.dx = Quantity(dx, xunit)
        new.xindex = xindex
        return new

    # -- series properties ----------------------

    @property
    def xindex(self):
        """Positions of the data on the x-axis

        :type: `~astropy.units.Quantity` array
        """
        try:
            return self._xindex
        except AttributeError:
            self._xindex = self.x0 + (
                numpy.arange(self.shape[0]) * self.dx)
            return self._xindex

    @xindex.setter
    def xindex(self, index):
        if index is None:
            del self.xindex
            return
        elif not isinstance(index, Quantity):
            index = Quantity(index, unit=self._default_xunit)
        self.x0 = index[0]
        if index.size:
            self.dx = index[1] - index[0]
        else:
            self.dx = None
        self._xindex = index

    @xindex.deleter
    def xindex(self):
        try:
            del self._xindex
        except AttributeError:
            pass

    @property
    def x0(self):
        """X-axis value of the first data point

        :type: `~astropy.units.Quantity` scalar
        """
        return self._x0

    @x0.setter
    def x0(self, value):
        if not isinstance(value, Quantity) and value is not None:
            try:
                value = Quantity(value, self._default_xunit)
            except TypeError:
                value = Quantity(float(value), self._default_xunit)
        try:
            x0 = self.x0
        except AttributeError:
            del self.xindex
        else:
            if value is None or self.x0 is None or value != x0:
                del self.xindex
        self._x0 = value

    @x0.deleter
    def x0(self):
        self._x0 = None

    @property
    def dx(self):
        """X-axis sample separation

        :type: `~astropy.units.Quantity` scalar
        """
        return self._dx

    @dx.setter
    def dx(self, value):
        if not isinstance(value, Quantity) and value is not None:
            value = Quantity(value).to(self.xunit)
        try:
            dx = self.dx
        except AttributeError:
            del self.xindex
        else:
            if value is None or self.dx is None or value != dx:
                del self.xindex
        self._dx = value

    @dx.deleter
    def dx(self):
        self._dx = None

    @property
    def xunit(self):
        """Unit of x-axis index

        :type: `~astropy.units.Unit`
        """
        return self.x0.unit

    @property
    def xspan(self):
        """X-axis [low, high) segment encompassed by these data

        :type: `~gwpy.segments.Segment`
        """
        from ..segments import Segment
        x0 = self.x0.to(self._default_xunit).value
        dx = self.dx.to(self._default_xunit).value
        return Segment(x0, x0+self.shape[0]*dx)

    # -- series methods -------------------------

    def copy(self, order='C'):
        new = super(Series, self).copy(order=order)
        try:
            new._xindex = self._xindex.copy()
        except AttributeError:
            pass
        return new
    copy.__doc__ = Array.copy.__doc__

    def zip(self):
        """Zip the `xindex` and `value` arrays of this `Series`

        Returns
        -------
        stacked : 2-d `numpy.ndarray`
            The array formed by stacking the the `xindex` and `value` of this
            series

        Examples
        --------
        >>> a = Series([0, 2, 4, 6, 8], xindex=[-5, -4, -3, -2, -1])
        >>> a.zip()
        array([[-5.,  0.],
               [-4.,  2.],
               [-3.,  4.],
               [-2.,  6.],
               [-1.,  8.]])

        """
        return numpy.column_stack((self.xindex.value, self.value))

    def diff(self, n=1, axis=-1):
        """Calculate the n-th order discrete difference along given axis.

        The first order difference is given by ``out[n] = a[n+1] - a[n]`` along
        the given axis, higher order differences are calculated by using `diff`
        recursively.

        Parameters
        ----------
        n : int, optional
            The number of times values are differenced.
        axis : int, optional
            The axis along which the difference is taken, default is the
            last axis.

        Returns
        -------
        diff : `Series`
            The `n` order differences. The shape of the output is the same
            as the input, except along `axis` where the dimension is
            smaller by `n`.

        See Also
        --------
        numpy.diff
            for documentation on the underlying method
        """
        out = super(Array, self).diff(n=n, axis=axis)
        out.x0 = self.x0 + self.dx * n
        return out

    diff.__doc__ = numpy.diff.__doc__

    def __array_finalize__(self, obj):
        super(Series, self).__array_finalize__(obj)
        if hasattr(self, '_xindex'):
            obj._xindex = self._xindex

    def __getslice__(self, i, j):
        new = super(Series, self).__getslice__(i, j)
        new.__dict__ = self.copy_metadata()
        new.x0 = self.x0 + i * self.dx
        return new

    def __getitem__(self, item):
        if isinstance(item, (float, int)):
            return Quantity(self.value[item], unit=self.unit)
        new = super(Series, self).__getitem__(item)
        if isinstance(item, slice):
            if item.start:
                new.x0 = self.x0 + item.start * self.dx
            if item.step:
                new.dx = self.dx * item.step
        elif isinstance(item, numpy.ndarray):
            new.xindex = self.xindex[item]
        return new

    # -- series manipulations ------------------

    def is_contiguous(self, other, tol=1/2.**18):
        """Check whether other is contiguous with self.

        Parameters
        ----------
        other : `Series`, `numpy.ndarray`
            another series of the same type to test for contiguity

        tol : `float`, optional
            the numerical tolerance of the test

        Returns
        -------
        1
            if `other` is contiguous with this series, i.e. would attach
            seamlessly onto the end
        -1
            if `other` is anti-contiguous with this seires, i.e. would attach
            seamlessly onto the start
        0
            if `other` is completely dis-contiguous with thie series

        Notes
        -----
        if a raw `numpy.ndarray` is passed as other, with no metadata, then
        the contiguity check will always pass
        """
        self.is_compatible(other)
        if isinstance(other, type(self)):
            if abs(float(self.xspan[1] - other.xspan[0])) < tol:
                return 1
            elif abs(float(other.xspan[1] - self.xspan[0])) < tol:
                return -1
            else:
                return 0
        elif type(other) in [list, tuple, numpy.ndarray]:
            return 1

    def is_compatible(self, other):
        """Check whether this series and other have compatible metadata

        This method tests that the `sample size <Series.dx>`, and the
        `~Series.unit` match.
        """
        if isinstance(other, type(self)):
            if not self.dx == other.dx:
                raise ValueError("%s sample sizes do not match: "
                                 "%s vs %s." % (type(self).__name__,
                                                self.dx, other.dx))
            if not self.unit == other.unit and not (
                    self.unit in [dimensionless_unscaled, None] and
                    other.unit in [dimensionless_unscaled, None]):
                raise ValueError("%s units do not match: %s vs %s."
                                 % (type(self).__name__, str(self.unit),
                                    str(other.unit)))
        else:
            arr = numpy.asarray(other)
            if arr.ndim != self.ndim:
                raise ValueError("Dimensionality does not match")
            if arr.dtype != self.dtype:
                warn("Array data types do not match: %s vs %s"
                     % (self.dtype, other.dtype))
        return True

    def append(self, other, gap='raise', inplace=True, pad=0.0, resize=True):
        """Connect another series onto the end of the current one.

        Parameters
        ----------
        other : `Series`
            another series of the same type to connect to this one

        gap : `str`, optional, default: ``'raise'``
            action to perform if there's a gap between the other series
            and this one. One of

                - ``'raise'`` - raise an `Exception`
                    - ``'ignore'`` - remove gap and join data
                - ``'pad'`` - pad gap with zeros

        inplace : `bool`, optional, default: `True`
            perform operation in-place, modifying current `Series`,
            otherwise copy data and return new `Series`

            .. warning::

               inplace append bypasses the reference check in
               `numpy.ndarray.resize`, so be carefully to only use this
               for arrays that haven't been sharing their memory!

        pad : `float`, optional, default: ``0.0``
            value with which to pad discontiguous series

        resize : `bool`, optional, default: `True`
            resize this array to accommodate new data, otherwise shift the
            old data to the left (potentially falling off the start) and
            put the new data in at the end

        Returns
        -------
        series : `Series`
            a new series containing joined data sets
        """
        # check metadata
        self.is_compatible(other)
        # make copy if needed
        if not inplace:
            self = self.copy()
        # fill gap
        if self.is_contiguous(other) != 1:
            if gap == 'pad':
                ngap = floor(
                    (other.xspan[0] - self.xspan[1]) / self.dx.value + 0.5)
                if ngap < 1:
                    raise ValueError(
                        "Cannot append {0} that starts before this one:\n"
                        "    {0} 1 span: {1}\n    {0} 2 span: {2}".format(
                            type(self).__name__, self.xspan, other.xspan))
                gapshape = list(self.shape)
                gapshape[0] = int(ngap)
                padding = numpy.ones(gapshape, dtype=self.dtype) * pad
                self.append(padding, inplace=True, resize=resize)
            elif gap == 'ignore':
                pass
            elif self.xspan[0] < other.xspan[0] < self.xspan[1]:
                raise ValueError(
                    "Cannot append overlapping {0}s:\n"
                    "    {0} 1 span: {1}\n    {0} 2 span: {2}".format(
                        type(self).__name__, self.xspan, other.xspan))
            else:
                raise ValueError(
                    "Cannot append discontiguous {0}\n"
                    "    {0} 1 span: {1}\n    {0} 2 span: {2}".format(
                        type(self).__name__, self.xspan, other.xspan))

        # check empty other
        if not other.size:
            return self

        # resize first
        if resize:
            N = other.shape[0]
            s = list(self.shape)
            s[0] = self.shape[0] + other.shape[0]
            try:
                self.resize(s, refcheck=False)
            except ValueError as e:
                if 'resize only works on single-segment arrays' in str(e):
                    self = self.copy()
                    self.resize(s)
                else:
                    raise
        elif other.shape[0] < self.shape[0]:
            N = other.shape[0]
            self.value[:-N] = self.value[N:]
        else:
            N = min(self.shape[0], other.shape[0])

        # if units are the same, can shortcut
        if type(other) == type(self) and other.unit == self.unit:
            self.value[-N:] = other.value[-N:]
        # otherwise if its just a numpy array
        elif type(other) == type(self.value) or (
                other.dtype.name.startswith('uint')):
            self.value[-N:] = other[-N:]
        else:
            self[-N:] = other[-N:]
        try:
            self._xindex
        except AttributeError:
            if not resize:
                self.x0 = self.x0.value + other.shape[0] * self.dx.value
        else:
            if resize:
                try:
                    self.xindex.resize((s[0],), refcheck=False)
                except ValueError as e:
                    if 'cannot resize' in str(e):
                        self.xindex = numpy.resize(self.xindex, (s[0],))
                    else:
                        raise
            else:
                self.xindex[:-other.shape[0]] = self.xindex[other.shape[0]:]
            try:
                self.xindex[-other.shape[0]:] = other._xindex
            except AttributeError:
                del self.xindex
            try:
                self.dx = self.xindex[1] - self.xindex[0]
            except IndexError:
                pass
            self.epoch = self.xindex[0]
        return self

    def prepend(self, other, gap='raise', inplace=True, pad=0.0, resize=True):
        """Connect another series onto the start of the current one.

        Parameters
        ----------
        other : `Series`
            another series of the same type as this one

        gap : `str`, optional, default: ``'raise'``
            action to perform if there's a gap between the other series
            and this one. One of

                - ``'raise'`` - raise an `Exception`
                - ``'ignore'`` - remove gap and join data
                - ``'pad'`` - pad gap with zeros

        inplace : `bool`, optional, default: `True`
            perform operation in-place, modifying current series,
            otherwise copy data and return new series

            .. warning::

               inplace prepend bypasses the reference check in
               `numpy.ndarray.resize`, so be carefully to only use this
               for arrays that haven't been sharing their memory!

        pad : `float`, optional, default: ``0.0``
            value with which to pad discontiguous `Series`
        resize : `bool`, optional, default: `True`

        Returns
        -------
        series : `TimeSeries`
            time-series containing joined data sets
        """
        out = other.append(self, gap=gap, inplace=False,
                           pad=pad, resize=resize)
        if inplace:
            self.resize(out.shape, refcheck=False)
            self[:] = out[:]
            self.x0 = out.x0.copy()
            del out
            return self
        else:
            return out

    def update(self, other, inplace=True):
        """Update this series by appending new data from an other
        and dropping the same amount of data off the start.

        This is a convenience method that just calls `~Series.append` with
        `resize=False`.
        """
        return self.append(other, inplace=inplace, resize=False)

    def crop(self, start=None, end=None, copy=False):
        """Crop this series to the given x-axis extent.

        Parameters
        ----------
        start : `float`, optional
            lower limit of x-axis to crop to, defaults to
            current `~Series.x0`

        end : `float`, optional
            upper limit of x-axis to crop to, defaults to current series end

        copy : `bool`, optional, default: `False`
            copy the input data to fresh memory, otherwise return a view

        Returns
        -------
        series : `Series`
            A new series with a sub-set of the input data

        Notes
        -----
        If either ``start`` or ``end`` are outside of the original
        `Series` span, warnings will be printed and the limits will
        be restricted to the :attr:`~Series.xspan`
        """
        # pin early starts to time-series start
        if start == self.xspan[0]:
            start = None
        elif start is not None and start < self.xspan[0]:
            warn('%s.crop given start smaller than current start, '
                 'crop will begin when the Series actually starts.'
                 % type(self).__name__)
            start = None
        # pin late ends to time-series end
        if end == self.xspan[1]:
            end = None
        if start is not None and end > self.xspan[1]:
            warn('%s.crop given end larger than current end, '
                 'crop will end when the Series actually ends.'
                 % type(self).__name__)
            end = None
        # find start index
        if start is None:
            idx0 = None
        else:
            idx0 = int(float(start - self.xspan[0]) / self.dx.value)
        # find end index
        if end is None:
            idx1 = None
        else:
            idx1 = int(float(end - self.xspan[0]) / self.dx.value)
            if idx1 >= self.size:
                idx1 = None
        # crop
        if copy:
            return self[idx0:idx1].copy()
        else:
            return self[idx0:idx1]

    def pad(self, pad_width, **kwargs):
        """Pad this series to a new size

        Parameters
        ----------
        pad_width : `int`, pair of `ints`
            number of samples by which to pad each end of the array.
            Single int to pad both ends by the same amount, or
            (before, after) `tuple` to give uneven padding
        **kwargs
            see :meth:`numpy.pad` for kwarg documentation

        Returns
        -------
        series : `Series`
            the padded version of the input

        See also
        --------
        numpy.pad
            for details on the underlying functionality
        """
        kwargs.setdefault('mode', 'constant')
        if isinstance(pad_width, int):
            pad_width = (pad_width,)
        new = numpy.pad(self.value, pad_width, **kwargs).view(type(self))
        new.__dict__ = self.copy_metadata()
        new.x0 -= self.dx * pad_width[0]
        return new



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""This module extends the functionality of the :mod:`glue.ligolw`
library for reading/writing/manipulating LIGO_LW format XML tables.

The :mod:`glue.ligolw.lsctables` library defines a number of specific
tables used in GW data analysis; this module extends their functionality
with the unified input/output system (the .read() method).

Additionally, for event tables (burst, inspiral, and ringdown), methods
to calculate event rate are also attached.

Users can make the extensions available by either importing the
:mod:`~glue.ligolw.lsctables` module from gwpy as follows::

    >>> from gwpy.table import lsctables

or simply importing the :mod:`gwpy.table` module at any point before
using the lsctables module (even after importing it from glue)::

    >>> from glue.ligolw import lsctables
    >>> import gwpy.table

"""

import warnings
warnings.filterwarnings('ignore', 'column name', UserWarning)

from glue.ligolw.ligolw import (Column, Document)
from glue.ligolw.table import Table

# import all tables
from . import lsctables


# attach unified I/O
from .io import *

# attach rate methods
from .rate import (event_rate, binned_event_rates)

from .. import version

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__credits__ = 'Kipp Cannon <kipp.cannon@ligo.org>'
__version__ = version.version
__all__ = ['Column', 'Document', 'Table', 'lsctables']



#!/usr/bin/env python

# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Plotting public LIGO data

I would like to study the gravitational wave strain time-series around the time of an interesting simulated signal during the last science run (S6).

These data are public, so we can load them directly from the web.
"""

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__currentmodule__ = 'gwpy.timeseries'

# First: import everything we need (and nothing we don't need)
from urllib2 import urlopen
from numpy import asarray
from gwpy.timeseries import TimeSeries

# Next, download the data as a string of text
data = urlopen('http://www.ligo.org/science/GW100916/L-strain_hp30-968654552-10.txt').read()

# We can now parse the text as a list of floats, and generate a `TimeSeries`
# by supplying the necessary metadata
ts = TimeSeries(asarray(data.splitlines(), dtype=float),
                epoch=968654552, sample_rate=16384, unit='strain')

# Finally, we can make a plot:
plot = ts.plot()
plot.set_title('LIGO Livingston Observatory data for GW100916')
plot.set_ylabel('Gravitational-wave strain amplitude')
plot.show()



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Shared GWF-file identifier
"""

from glue.lal import CacheEntry

from astropy.io import registry

from ... import (TimeSeries, TimeSeriesDict, StateVector, StateVectorDict)
from .... import version

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__version__ = version.version


def identify_gwf(*args, **kwargs):
    """Determine an input file as written in GWF-format.
    """
    filename = args[1]
    ce = args[3]
    if isinstance(ce, CacheEntry):
        filename = ce.path
    if isinstance(filename, str) and filename.endswith('gwf'):
        return True
    else:
        return False

registry.register_identifier('gwf', TimeSeries, identify_gwf)
registry.register_identifier('gwf', TimeSeriesDict, identify_gwf)
registry.register_identifier('gwf', StateVector, identify_gwf)
registry.register_identifier('gwf', StateVectorDict, identify_gwf)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""An extension of the Plot class for handling TimeSeries
"""

import re
import datetime
import numpy
import copy

from matplotlib import (pyplot, cm, colors)
from matplotlib.projections import register_projection
from matplotlib.artist import allow_rasterization
from matplotlib.cbook import iterable

try:
    from mpl_toolkits.axes_grid1 import make_axes_locatable
except ImportError:
    from mpl_toolkits.axes_grid import make_axes_locatable


from ..time import LIGOTimeGPS
from . import (gps, tex)
from .core import Plot
from ..segments import SegmentList
from ..time import Time
from .axes import Axes
from .decorators import auto_refresh

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__all__ = ['TimeSeriesPlot', 'TimeSeriesAxes']


class TimeSeriesAxes(Axes):
    """Custom `Axes` for a :class:`~gwpy.plotter.TimeSeriesPlot`.
    """
    name = 'timeseries'

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('xscale', 'auto-gps')
        super(TimeSeriesAxes, self).__init__(*args, **kwargs)
        self.fmt_xdata = lambda t: LIGOTimeGPS(t)
        self.set_xlabel('_auto')

    @allow_rasterization
    def draw(self, *args, **kwargs):
        # dynamically set scaling
        if self.get_xscale() == 'auto-gps':
            self.auto_gps_scale()
        # dynamically set x-axis label
        nolabel = self.get_xlabel() == '_auto'
        if nolabel:
            self.auto_gps_label()
        # auto-detect GPS scales
        super(TimeSeriesAxes, self).draw(*args, **kwargs)
        # reset label
        if nolabel:
            self.set_xlabel('_auto')

    draw.__doc__ = Axes.draw.__doc__

    # -----------------------------------------------
    # GPS scaling

    def set_xscale(self, scale):
        super(TimeSeriesAxes, self).set_xscale(scale)
        if scale != 'auto-gps' and self.get_xlabel() == '_auto':
            self.set_xlabel('')

    def auto_gps_label(self):
        scale = self.xaxis._scale
        epoch = scale.get_epoch()
        if epoch is None:
            self.set_xlabel('GPS Time')
        else:
            unit = scale.get_unit_name()
            utc = re.sub('\.0+', '',
                         Time(epoch, format='gps', scale='utc').iso)
            self.set_xlabel('Time [%s] from %s UTC (%s)' % (unit, utc, epoch))

    def auto_gps_scale(self):
        """Automagically set the GPS scale for the time-axis of this plot
        based on the current view limits
        """
        self.xaxis._set_scale('auto-gps', epoch=self.get_epoch())

    def set_epoch(self, epoch):
        try:
            xscale = self.get_xscale()
        except AttributeError:
            pass
        else:
            self.xaxis._set_scale(xscale, epoch=epoch)

    def get_epoch(self):
        return self.xaxis._scale.get_epoch()

    epoch = property(fget=get_epoch, fset=set_epoch, doc=get_epoch.__doc__)

    def set_xlim(self, left=None, right=None, emit=True, auto=False, **kw):
        if right is None and iterable(left):
            left, right = left
        left = float(left)
        right = float(right)
        if 'gps' in self.get_xscale() and self.epoch is None:
            self.set_epoch(left)
        super(TimeSeriesAxes, self).set_xlim(left=left, right=right, emit=emit,
                                             auto=auto, **kw)
    set_xlim.__doc__ = Axes.set_xlim.__doc__

    # -------------------------------------------
    # GWpy class plotting methods

    @auto_refresh
    def plot(self, *args, **kwargs):
        """Plot data onto these Axes.

        Parameters
        ----------
        args
            a single :class:`~gwpy.timeseries.TimeSeries` (or sub-class)
            or standard (x, y) data arrays
        kwargs
            keyword arguments applicable to :meth:`~matplotib.axens.Axes.plot`

        Returns
        -------
        Line2D
            the :class:`~matplotlib.lines.Line2D` for this line layer

        See Also
        --------
        :meth:`matplotlib.axes.Axes.plot`
            for a full description of acceptable ``*args` and ``**kwargs``
        """
        from ..timeseries import TimeSeries
        from ..spectrogram import Spectrogram
        if len(args) == 1 and isinstance(args[0], TimeSeries):
            return self.plot_timeseries(*args, **kwargs)
        elif len(args) == 1 and isinstance(args[0], Spectrogram):
            return self.plot_spectrogram(*args, **kwargs)
        else:
            return super(TimeSeriesAxes, self).plot(*args, **kwargs)

    @auto_refresh
    def plot_timeseries(self, timeseries, **kwargs):
        """Plot a :class:`~gwpy.timeseries.TimeSeries` onto these
        axes

        Parameters
        ----------
        timeseries : :class:`~gwpy.timeseries.TimeSeries`
            data to plot
        **kwargs
            any other keyword arguments acceptable for
            :meth:`~matplotlib.Axes.plot`

        Returns
        -------
        Line2D
            the :class:`~matplotlib.lines.Line2D` for this line layer

        See Also
        --------
        :meth:`matplotlib.axes.Axes.plot`
            for a full description of acceptable ``*args` and ``**kwargs``
        """
        if tex.USE_TEX:
            kwargs.setdefault('label', tex.label_to_latex(timeseries.name))
        else:
            kwargs.setdefault('label', timeseries.name)
        if not self.epoch:
            self.set_epoch(timeseries.x0)
        line = self.plot(timeseries.times.value, timeseries.value, **kwargs)
        if len(self.lines) == 1 and timeseries.size:
            self.set_xlim(*timeseries.span)
        return line

    @auto_refresh
    def plot_timeseries_mmm(self, mean_, min_=None, max_=None, **kwargs):
        """Plot a `TimeSeries` onto these axes, with (min, max) shaded
        regions

        The `mean_` `TimeSeries` is plotted normally, while the `min_`
        and `max_ `TimeSeries` are plotted lightly below and above,
        with a fill between them and the mean_.

        Parameters
        ----------
        mean_ : :class:`~gwpy.timeseries.TimeSeries`
            data to plot normally
        min_ : :class:`~gwpy.timeseries.TimeSeries`
            first data set to shade to mean_
        max_ : :class:`~gwpy.timeseries.TimeSeries`
            second data set to shade to mean_
        **kwargs
            any other keyword arguments acceptable for
            :meth:`~matplotlib.Axes.plot`

        Returns
        -------
        artists : `tuple`
            a 5-tuple containing (Line2d for mean_, `Line2D` for min_,
            `PolyCollection` for min_ shading, `Line2D` for max_, and
            `PolyCollection` for max_ shading)

        See Also
        --------
        :meth:`matplotlib.axes.Axes.plot`
            for a full description of acceptable ``*args` and ``**kwargs``
        """
        # plot mean
        line1 = self.plot_timeseries(mean_, **kwargs)[0]
        # plot min and max
        kwargs.pop('label', None)
        color = kwargs.pop('color', line1.get_color())
        linewidth = kwargs.pop('linewidth', line1.get_linewidth()) / 2
        if min_ is not None:
            a = self.plot(min_.times.value, min_.value, color=color,
                          linewidth=linewidth, **kwargs)
            b = self.fill_between(min_.times.value, mean_.value, min_.value,
                                  alpha=0.1, color=color)
        else:
            a = b = None
        if max_ is not None:
            c = self.plot(max_.times.value, max_.value, color=color,
                          linewidth=linewidth, **kwargs)
            d = self.fill_between(max_.times.value, mean_.value, max_.value,
                                  alpha=0.1, color=color)
        else:
            c = d = None
        return line1, a, b, c, d

    @auto_refresh
    def plot_spectrogram(self, spectrogram, **kwargs):
        """Plot a :class:`~gwpy.spectrogram.core.Spectrogram` onto
        these axes

        Parameters
        ----------
        spectrogram : :class:`~gwpy.spectrogram.core.Spectrogram`
            data to plot
        **kwargs
            any other keyword arguments acceptable for
            :meth:`~matplotlib.Axes.plot`

        Returns
        -------
        Line2D
            the :class:`~matplotlib.lines.Line2D` for this line layer

        See Also
        --------
        :meth:`matplotlib.axes.Axes.plot`
            for a full description of acceptable ``*args` and ``**kwargs``
        """
        # rescue grid settings
        grid = (self.xaxis._gridOnMajor, self.xaxis._gridOnMinor,
                self.yaxis._gridOnMajor, self.yaxis._gridOnMinor)

        norm = kwargs.pop('norm', None)
        if norm == 'log':
            vmin = kwargs.get('vmin', None)
            vmax = kwargs.get('vmax', None)
            norm = colors.LogNorm(vmin=vmin, vmax=vmax)
        kwargs['norm'] = norm
        if not self.epoch:
            self.set_epoch(spectrogram.x0)
        x = numpy.concatenate((spectrogram.times.value,
                               [spectrogram.span[-1]]))
        y = numpy.concatenate((spectrogram.frequencies.value,
                               [spectrogram.band[-1]]))
        X, Y = numpy.meshgrid(x, y, copy=False, sparse=True)
        mesh = self.pcolormesh(X, Y, spectrogram.value.T, **kwargs)
        if len(self.collections) == 1:
            self.set_xlim(*spectrogram.span)
            self.set_ylim(*spectrogram.band)
        if not self.get_ylabel():
            self.add_label_unit(spectrogram.yunit, axis='y')

        # reset grid
        if grid[0]:
            self.xaxis.grid(True, 'major')
        if grid[1]:
            self.xaxis.grid(True, 'minor')
        if grid[2]:
            self.yaxis.grid(True, 'major')
        if grid[3]:
            self.yaxis.grid(True, 'minor')
        return mesh

register_projection(TimeSeriesAxes)


class TimeSeriesPlot(Plot):
    """`Figure` for displaying a :class:`~gwpy.timeseries.TimeSeries`.

    Parameters
    ----------
    *series : `TimeSeries`
        any number of :class:`~gwpy.timeseries.TimeSeries` to
        display on the plot
    **kwargs
        other keyword arguments as applicable for the
        :class:`~gwpy.plotter.Plot`
    """
    _DefaultAxesClass = TimeSeriesAxes

    def __init__(self, *series, **kwargs):
        """Initialise a new TimeSeriesPlot
        """
        kwargs.setdefault('projection', self._DefaultAxesClass.name)
        kwargs.setdefault('figsize', [12, 6])
        # extract custom keyword arguments
        sep = kwargs.pop('sep', False)
        epoch = kwargs.pop('epoch', None)
        sharex = kwargs.pop('sharex', False)
        sharey = kwargs.pop('sharey', False)
        # separate keyword arguments
        axargs, plotargs = self._parse_kwargs(kwargs)

        # generate figure
        super(TimeSeriesPlot, self).__init__(**kwargs)

        # plot data
        axesdata = self._get_axes_data(series, sep=sep)
        for data in axesdata:
            ax = self._add_new_axes(**axargs)
            for ts in data:
                ax.plot(ts, **plotargs)
            if 'sharex' not in axargs and sharex is True:
                axargs['sharex'] = ax
            if 'sharey' not in axargs and sharey is True:
                axargs['sharey'] = ax
        axargs.pop('sharex', None)
        axargs.pop('sharey', None)
        axargs.pop('projection', None)

        # set epoch
        if len(self.axes):
            flatdata = [ts for data in axesdata for ts in data]
            span = SegmentList([ts.span for ts in flatdata]).extent()
            for ax in self.axes:
                for key, val in axargs.iteritems():
                    getattr(ax, 'set_%s' % key)(val)
                if epoch is not None:
                    ax.set_epoch(epoch)
                else:
                    ax.set_epoch(span[0])
                if 'xlim' not in axargs:
                    ax.set_xlim(*span)
                if 'label' in plotargs:
                    ax.legend()
            for ax in self.axes[:-1]:
                ax.set_xlabel("")

    # -----------------------------------------------
    # properties

    @property
    def epoch(self):
        """Find the GPS epoch of this plot
        """
        axes = self._find_axes(self._DefaultAxesClass.name)
        return axes.epoch

    def get_epoch(self):
        return self.epoch

    @auto_refresh
    def set_epoch(self, gps):
        """Set the GPS epoch of this plot
        """
        axeslist = self.get_axes(self._DefaultAxesClass.name)
        for axes in axeslist:
            axes.set_epoch(gps)

    # -----------------------------------------------
    # TimeSeriesPlot methods

    def add_timeseries(self, timeseries, **kwargs):
        super(TimeSeriesPlot, self).add_timeseries(timeseries, **kwargs)
        if self.epoch is None:
            self.set_epoch(timeseries.epoch)

    def add_state_segments(self, segments, ax=None, height=0.2, pad=0.1,
                           location='bottom', plotargs=dict()):
        """Add a `SegmentList` to this `TimeSeriesPlot` indicating state
        information about the main Axes data.

        By default, segments are displayed in a thin horizontal set of Axes
        sitting immediately below the x-axis of the main

        Parameters
        ----------
        segments : :class:`~gwpy.segments.flag.DataQualityFlag`
            A data-quality flag, or `SegmentList` denoting state segments
            about this Plot
        ax : `Axes`
            specific Axes set against which to anchor new segment Axes
        plotargs
            keyword arguments passed to
            :meth:`~gwpy.plotter.SegmentAxes.plot`
        """
        from .segments import SegmentAxes
        if not ax:
            try:
                ax = self.get_axes(self._DefaultAxesClass.name)[-1]
            except IndexError:
                raise ValueError("No 'timeseries' Axes found, cannot anchor "
                                 "new segment Axes.")
        pyplot.setp(ax.get_xticklabels(), visible=False)
        # add new axes
        if ax.get_axes_locator():
            divider = ax.get_axes_locator()._axes_divider
        else:
            divider = make_axes_locatable(ax)
        if location not in ['top', 'bottom']:
            raise ValueError("Segments can only be positoned at 'top' or "
                             "'bottom'.")
        segax = divider.append_axes(location, height, pad=pad,
                                    axes_class=SegmentAxes, sharex=ax)
        segax.set_xscale(ax.get_xscale())

        # plot segments and set axes properties
        segax.plot(segments, **plotargs)
        segax.grid(b=False, which='both', axis='y')
        segax.autoscale(axis='y', tight=True)
        # set ticks and label
        segax.set_xlabel(ax.get_xlabel())
        ax.set_xlabel("")
        segax.set_xlim(*ax.get_xlim())
        return segax



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Decorator for GWpy plotting
"""

from functools import wraps

from matplotlib.figure import Figure
from matplotlib.axes import Axes

from .. import version
__version__ = version.version
__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


def auto_refresh(func):
    """Decorate `func` to refresh the containing figure when complete
    """
    @wraps(func)
    def wrapper(artist, *args, **kwargs):
        """Call the method, and refresh the figure on exit
        """
        refresh = kwargs.pop('refresh', False)
        try:
            return func(artist, *args, **kwargs)
        finally:
            if isinstance(artist, Axes):
                if refresh or artist.figure.get_auto_refresh():
                    artist.figure.refresh()
            elif isinstance(artist, Figure):
                if refresh or artist.get_auto_refresh():
                    artist.refresh()
            else:
                raise TypeError("Cannot determine containing Figure for "
                                "auto_refresh() decorator")
    return wrapper


def axes_method(func):
    """Decorate `func` to call the same method of the contained `Axes`

    Raises
    ------
    RuntimeError
        if multiple `Axes` are found when the method is called.
        This method makes no attempt to decide which `Axes` to use
    """
    @wraps(func)
    def wrapper(figure, *args, **kwargs):
        """Find the relevant `Axes` and call the method
        """
        axes = [ax for ax in figure.axes if ax not in figure._coloraxes]
        if len(axes) == 0:
            raise RuntimeError("No axes found for which '%s' is applicable"
                               % func.__name__)
        if len(axes) != 1:
            raise RuntimeError("{0} only applicable for a Plot with a "
                               "single set of data axes. With multiple "
                               "data axes, you should access the {0} "
                               "method of the relevant Axes (stored in "
                               "``Plot.axes``) directly".format(func.__name__))
        axesf = getattr(axes[0], func.__name__)
        return axesf(*args, **kwargs)
    return wrapper



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""This module provides a spectral-variation histogram class
"""

import numpy
import sys

if sys.version_info[0] < 3:
    range = xrange

from .. import version
__version__ = version.version
__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'

from ..data import (Quantity, Array, Array2D)
from .core import Spectrum
from ..detector import Channel

__all__ = ['SpectralVariance']


class SpectralVariance(Array2D):
    """A 2-dimensional array containing the variance histogram of a
    frequency-series `Spectrum`
    """
    _metadata_slots = Spectrum._metadata_slots + ['bins']
    xunit = Spectrum.xunit

    def __new__(cls, data, bins, name=None, channel=None, epoch=None, unit=None,
                f0=None, df=None, **kwargs):
        """Generate a new SpectralVariance
        """
        # parse Channel input
        if channel:
            channel = Channel(channel)
            name = name or channel.name
            unit = unit or channel.unit
        # generate Spectrogram
        new = super(SpectralVariance, cls).__new__(
            cls, data, name=name, channel=channel, epoch=epoch,
            x0=f0, dx=df, y0=None, dy=None, **kwargs)
        new.bins = bins
        return new

    # -------------------------------------------
    # SpectralVariance properties

    @property
    def normed(self):
        return self._normed

    @property
    def density(self):
        return self._density

    @property
    def bins(self):
        return self._bins

    @bins.setter
    def bins(self, bins):
        if bins is None:
            del self.bins
            return
        bins = Quantity(bins)
        if bins.size != self.shape[1] + 1:
            raise ValueError(
                "SpectralVariance.bins must be given as a list of bin edges, "
                "including the rightmost edge, and have length 1 greater than "
                "the y-axis of the SpectralVariance data")
        self._bins = bins

    @bins.deleter
    def bins(self):
        try:
            del self._bins
        except AttributeError:
            pass

    # over-write yindex to communicate with bins
    @property
    def yindex(self):
        return self.bins[:-1]

    @property
    def dy(self):
        return self.bins[1] - self.bins[0]

    @property
    def y0(self):
        return self.bins[0]

    f0 = property(Array2D.x0.__get__, Array2D.x0.__set__,
                  Array2D.x0.__delete__,
                  """Starting frequency for this `Spectrogram`

                  This attributes is recorded as a
                  :class:`~astropy.units.quantity.Quantity` object, assuming a
                  unit of 'Hertz'.
                  """)

    df = property(Array2D.dx.__get__, Array2D.dx.__set__,
                  Array2D.dx.__delete__,
                  """Frequency spacing of this `Spectogram`

                  This attributes is recorded as a
                  :class:`~astropy.units.quantity.Quantity` object, assuming a
                  unit of 'Hertz'.
                  """)

    frequencies = property(fget=Array2D.xindex.__get__,
                           fset=Array2D.xindex.__set__,
                           fdel=Array2D.xindex.__delete__,
                           doc="""Array of frequencies for each sample""")

    # -------------------------------------------
    # SpectralVariance methods

    @classmethod
    def from_spectrogram(cls, *spectrograms, **kwargs):
        """Calculate a new `SpectralVariance` from a
        :class:`~gwpy.spectrogram.core.Spectrogram`

        Parameters
        ----------
        spectrogram : :class:`~gwpy.spectrogram.core.Spectrogram`
            input `Spectrogram` data
        bins : :class:`~numpy.ndarray`, optional, default `None`
            array of histogram bin edges, including the rightmost edge
        low : `float`, optional, default: `None`
            left edge of lowest amplitude bin, only read
            if ``bins`` is not given
        high : `float`, optional, default: `None`
            right edge of highest amplitude bin, only read
            if ``bins`` is not given
        nbins : `int`, optional, default: `500`
            number of bins to generate, only read if ``bins`` is not
            given
        log : `bool`, optional, default: `False`
            calculate amplitude bins over a logarithmic scale, only
            read if ``bins`` is not given
        norm : `bool`, optional, default: `False`
            normalise bin counts to a unit sum
        density : `bool`, optional, default: `False`
            normalise bin counts to a unit integral

        Returns
        -------
        specvar : `SpectralVariance`
            2D-array of spectral frequency-amplitude counts

        See Also
        --------
        :func:`numpy.histogram`
            for details on specifying bins and weights
        """
        # parse args and kwargs
        if not len(spectrograms):
            raise ValueError("Must give at least one Spectrogram")
        bins = kwargs.pop('bins', None)
        low = kwargs.pop('low', None)
        high = kwargs.pop('high', None)
        nbins = kwargs.pop('nbins', 500)
        log = kwargs.pop('log', False)
        norm = kwargs.pop('norm', False)
        density = kwargs.pop('density', False)
        if norm and density:
            raise ValueError("Cannot give both norm=True and density=True, "
                             "please pick one")

        # get data and bins
        spectrogram = spectrograms[0]
        data = numpy.vstack(s.value for s in spectrograms)
        ubins = (bins is not None)
        if bins is None:
            if low is None and log:
                low = numpy.log10(data.min() / 2)
            elif low is None:
                low = data.min()/2
            elif log:
                low = numpy.log10(low)
            if high is None and log:
                high = numpy.log10(data.max() * 2)
            elif high is None:
                high = data.max() * 2
            elif log:
                high = numpy.log10(high)
            if log:
                bins = numpy.logspace(low, high, num=nbins+1)
            else:
                bins = numpy.linspace(low, high, num=nbins+1)
        nbins = bins.size-1
        bins = bins * spectrogram.unit

        # loop over frequencies
        out = numpy.zeros((data.shape[1], nbins))
        for i in range(data.shape[1]):
            out[i, :], bins = numpy.histogram(data[:, i], bins,
                                              density=density)
            if norm and out[i, :].sum():
                out[i, :] /= out[i, :].sum()

        # return SpectralVariance
        name = '%s variance' % spectrogram.name
        new = cls(out, bins, epoch=spectrogram.epoch, name=name,
                  channel=spectrogram.channel, f0=spectrogram.f0,
                  df=spectrogram.df)
        new._normed = norm
        new._density = density
        return new

    def percentile(self, percentile):
        """Calculate a given spectral percentile for this `SpectralVariance`

        Parameters
        ----------
        percentile : `float`
            percentile (0 - 100) of the bins to compute

        Returns
        -------
        spectrum : :class:`~gwpy.spectrum.core.Spectrum`
            the given percentile `Spectrum` calculated from this
            `SpectralVaraicence`
        """
        rows, columns = self.shape
        out = numpy.zeros(rows)
        # Loop over frequencies
        for i in range(rows):
            # Calculate cumulative sum for array
            cumsumvals = numpy.cumsum(self.value[i, :])

            # Find value nearest requested percentile
            abs_cumsumvals_minus_percentile = numpy.abs(cumsumvals -
                                                        percentile)
            minindex = abs_cumsumvals_minus_percentile.argmin()
            val = self.bins[minindex]
            out[i] = val
        name = '%s %s%% percentile' % (self.name, percentile)
        return Spectrum(out, epoch=self.epoch, frequencies=self.bins[:-1],
                        channel=self.channel, name=name)

    def plot(self, **kwargs):
        """Plot this `SpectralVariance`.
        """
        from ..plotter import SpectrumPlot
        return SpectrumPlot(self, **kwargs)



#!/usr/bin/env python

# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Calculating the coherence between two channels

The `coherence <http://en.wikipedia.org/wiki/Coherence_(physics)>`_ between
two channels is a measure of the frequency-domain correlation between their
time-series data.

In LIGO, the coherence is a crucial indicator of how noise sources couple into
the main differential arm-length readout.

In this example we calculate the coherence between two length-sensing
degrees of freedom, the Signal-Recycling Cavity length (SRCL), and the
Common-Arm motion (CARM).

"""

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__currentmodule__ = 'gwpy.timeseries'

# First, we import the `TimeSeriesDict`
from gwpy.timeseries import TimeSeriesDict

# and then :meth:`~TimeSeriesDict.get` both data sets:
data = TimeSeriesDict.get(['L1:LSC-SRCL_IN1_DQ', 'L1:LSC-CARM_IN1_DQ'],
                           'Feb 13 2015', 'Feb 13 2015 00:15')

# We can then calculate the :meth:`~TimeSeries.coherence` of one
# `TimeSeries` with respect to the other, using an 8-second Fourier
# transform length, with a 4-second (50%) overlap:
coh = data['L1:LSC-SRCL_IN1_DQ'].coherence(data['L1:LSC-CARM_IN1_DQ'], 8, 4)

# Finally, we can :meth:`~gwpy.spectrum.Spectrum.plot` the resulting data:
plot = coh.plot(figsize=[12, 6], label=None)
ax = plot.gca()
ax.set_yscale('linear')
ax.set_xlabel('Frequency [Hz]')
ax.set_ylabel('Coherence')
ax.set_title('Coherence between SRCL and CARM for L1')
ax.grid(True, 'both', 'both')
plot.show()



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""This module provides a few utilities for handling optional
dependencies within GWpy code.
"""

import inspect
from functools import wraps

from .. import version
__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__version__ = version.version


def import_method_dependency(module, stacklevel=1):
    """Import the given module, with a more useful `ImportError` message.

    Parameters
    ----------
    module : `str`
        name of the module to import

    Returns
    -------
    object : `object`
        the imported `object`

    Raises
    ------
    ImportError
        if the given object cannot be imported
    """
    try:
        return __import__(module, fromlist=[''])
    except ImportError as e:
        # get method name
        caller = inspect.stack()[stacklevel][3]
        # raise a better exception
        if not e.args:
            e.args = ('',)
        e.args = ("Cannot import %s required by the %s() method: %r"
                  % (module, caller, str(e)),) + e.args[1:]
        raise


def with_import(module):
    """Decorate a given method to import an optional dependency.

    Parameters
    ----------
    module : `str`
        name of the module object to import

    Returns
    -------
    decorator : `function`
        a decorator to apply to a method with the optional import
    """
    modname = module.split('.')[-1]

    def decorate_method(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            func.func_globals[modname] = import_method_dependency(module,
                                                                  stacklevel=2)
            return func(*args, **kwargs)
        return wrapper
    return decorate_method



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""This module defines the `Channel` and `ChannelList` classes.
"""

from __future__ import print_function

import re
import numpy
import warnings
import subprocess
import sys
from math import log

from astropy import units

try:
    from ..io.nds import (NDS2_CHANNEL_TYPE, NDS2_CHANNEL_TYPESTR)
except ImportError:
    NDS2_CHANNEL_TYPESTR = {
        1: 'online',
         2: 'raw',
         4: 'reduced',
         8: 's-trend',
         16: 'm-trend',
         32: 'test-pt',
         64: 'static',
         128: 'rds',
    }
    NDS2_CHANNEL_TYPE = dict((val, key) for (key, val) in
                             NDS2_CHANNEL_TYPESTR.iteritems())

from .. import version
from ..io import (reader, writer, datafind)
from ..time import to_gps
from ..utils.deps import with_import

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__version__ = version.version

re_quote = re.compile(r'^[\s\"\']+|[\s\"\']+$')


class Channel(object):
    """Representation of a laser-interferometer data channel.

    Parameters
    ----------
    name : `str`, `Channel`
        name of this Channel (or another  Channel itself).
        If a `Channel` is given, all other parameters not set explicitly
        will be copied over.
    sample_rate : `float`, `~astropy.units.Quantity`, optional
        number of samples per second
    unit : `~astropy.units.Unit`, `str`, optional
        name of the unit for the data of this channel
    frequency_range : `tuple` of `float`
        [low, high) spectral frequency range of interest for this channel
    safe : `bool`, optional
        is this channel 'safe' to use as a witness of non-gravitational-wave
        noise in the interferometer
    dtype : `numpy.dtype`, optional
        numeric type of data for this channel
    frametype : `str`, optional
        LDAS name for frames that contain this channel
    model : `str`, optional
        name of the SIMULINK front-end model that produces this channel

    Notes
    -----
    The `Channel` structure implemented here is designed to match the
    data recorded in the LIGO Channel Information System
    (https://cis.ligo.org) for which a query interface is provided.
    """
    MATCH = re.compile(
        r'((?:(?P<ifo>[A-Z]\d))?|[\w-]+):'  # match IFO prefix
         '(?:(?P<system>[a-zA-Z0-9]+))?'  # match system
         '(?:[-_](?P<subsystem>[a-zA-Z0-9]+))?'  # match subsystem
         '(?:_(?P<signal>[a-zA-Z0-9_]+))?'  # match signal
         '(?:\.(?P<trend>[a-z]+))?'  # match trend type
         '(?:,(?P<type>([a-z]-)?[a-z]+))?'  # match channel type
    )

    def __init__(self, name, sample_rate=None, unit=None, frequency_range=None,
                 safe=None, type=None, dtype=None, frametype=None,
                 model=None, url=None):
        """Create a new `Channel`
        """
        # copy existing Channel
        if isinstance(name, Channel):
            sample_rate = sample_rate or name.sample_rate
            unit = unit or name.unit
            frequency_range = frequency_range or name.frequency_range
            safe = safe or name.safe
            type = type or name.type
            dtype = dtype or name.dtype
            frametype = frametype or name.frametype
            model = model or name.model
            url = url or name.url
            name = str(name)
        # make a new channel
        # strip off NDS stuff for 'name'
        # parse name into component parts
        try:
            parts = self.parse_channel_name(name)
        except (TypeError, ValueError):
            self.name = str(name)
        else:
            self.name = str(name).split(',')[0]
            for key, val in parts.iteritems():
                try:
                    setattr(self, key, val)
                except AttributeError:
                    setattr(self, '_%s' % key, val)
        # set metadata
        if type is not None:
            self.type = type
        self.sample_rate = sample_rate
        self.unit = unit
        self.frequency_range = frequency_range
        self.safe = safe
        self.dtype = dtype
        self.frametype = frametype
        self.model = model
        self.url = url

    # -------------------------------------------------------------------------
    # read-write properties

    @property
    def name(self):
        """Name of this channel.

        This should follow the naming convention, with the following
        format: 'IFO:SYSTEM-SUBSYSTEM_SIGNAL'

        :type: `str`
        """
        try:
            return self._name
        except AttributeError:
            self._name = None

    @name.setter
    def name(self, n):
        self._name = n

    @property
    def sample_rate(self):
        """Rate of samples (Hertz) for this channel.

        :type: `~astropy.units.Quantity`
        """
        try:
            return self._sample_rate
        except AttributeError:
            self._sample_rate = None

    @sample_rate.setter
    def sample_rate(self, rate):
        if isinstance(rate, units.Unit):
            self._sample_rate = rate
        elif rate is None:
            self._sample_rate = None
        elif isinstance(rate, units.Quantity):
            self._sample_rate = rate
        else:
            self._sample_rate = units.Quantity(float(rate), unit=units.Hertz)

    @property
    def unit(self):
        """Data unit for this channel.

        :type: `~astropy.units.Unit`
        """
        return self._unit

    @unit.setter
    def unit(self, u):
        if u is None:
            self._unit = None
        else:
            self._unit = units.Unit(u)

    @property
    def frequency_range(self):
        """Frequency range of interest (Hertz) for this channel

        :type: `~astropy.units.Quantity` array
        """
        return self._frequency_range

    @frequency_range.setter
    def frequency_range(self, frange):
        if frange is None:
            del self.frequency_range
        else:
            low, hi = frange
            self._frequency_range = units.Quantity((low, hi), unit='Hz')

    @frequency_range.deleter
    def frequency_range(self):
        self._frequency_range = None

    @property
    def safe(self):
        """Whether this channel is 'safe' to use as a noise witness

        Any channel that records part or all of a GW signal as it
        interacts with the interferometer is not safe to use as a noise
        witness

        A safe value of `None` simply indicates that the safety of this
        channel has not been determined

        :type: `bool` or `None`
        """
        return self._safe

    @safe.setter
    def safe(self, s):
        if s is None:
            self._safe = None
        else:
            self._safe = bool(s)

    @property
    def model(self):
        """Name of the SIMULINK front-end model that defines this channel.

        :type: `str`
        """
        return self._model

    @model.setter
    def model(self, mdl):
        self._model = mdl and mdl.lower() or mdl

    @property
    def type(self):
        """DAQ data type for this channel.

        Valid values for this field are restricted to those understood
        by the NDS2 client sofware, namely:

        'm-trend', 'online', 'raw', 'reduced', 's-trend', 'static', 'test-pt'

        :type: `str`
        """
        try:
            return self._type
        except AttributeError:
            self._type = None
            return self.type

    @type.setter
    def type(self, type_):
        if type_ is None:
            self._type = None
        elif type_ in NDS2_CHANNEL_TYPESTR:
            self._type = NDS2_CHANNEL_TYPESTR[type_]
        elif type_.lower() in NDS2_CHANNEL_TYPE:
            self._type = type_.lower()
        else:
            raise ValueError("Channel type '%s' not understood." % type_)

    @property
    def ndstype(self):
        """NDS type integer for this channel.

        This property is mapped to the `Channel.type` string.
        """
        return NDS2_CHANNEL_TYPE.get(self.type)

    @ndstype.setter
    def ndstype(self, type_):
        self.type = type_

    @property
    def dtype(self):
        """Numeric type for data in this channel.

        :type: `~numpy.dtype`
        """
        return self._dtype

    @dtype.setter
    def dtype(self, type_):
        if type_ is None:
            self._dtype = None
        else:
            self._dtype = numpy.dtype(type_)

    @property
    def url(self):
        """CIS browser url for this channel.

        :type: `str`
        """
        return self._url

    @url.setter
    def url(self, href):
        self._url = href

    @property
    def frametype(self):
        """LDAS type description for frame files containing this channel.
        """
        return self._frametype

    @frametype.setter
    def frametype(self, ft):
        self._frametype = ft

    # -------------------------------------------------------------------------
    # read-only properties

    @property
    def ifo(self):
        """Interferometer prefix for this channel.

        :type: `str`
        """
        try:
            return self._ifo
        except AttributeError:
            self._ifo = None

    @property
    def system(self):
        """Instrumental system for this channel.

        :type: `str`
        """
        try:
            return self._system
        except AttributeError:
            self._system = None

    @property
    def subsystem(self):
        """Instrumental sub-system for this channel.

        :type: `str`
        """
        try:
            return self._subsystem
        except AttributeError:
            self._subsystem = None

    @property
    def signal(self):
        """Instrumental signal for this channel.

        :type: `str`
        """
        try:
            return self._signal
        except AttributeError:
            self._signal = None

    @property
    def trend(self):
        """Trend type for this channel.

        :type: `str`
        """
        try:
            return self._trend
        except AttributeError:
            self._trend = None

    @property
    def texname(self):
        """Name of this channel in LaTeX printable format.
        """
        return str(self).replace("_", r"\_")

    @property
    def ndsname(self):
        """Name of this channel as stored in the NDS database
        """
        if self.type not in [None, 'raw', 'reduced']:
            return '%s,%s' % (self.name, self.type)
        return self.name

    # -------------------------------------------------------------------------
    # classsmethods

    @classmethod
    def query(cls, name, debug=False, timeout=None):
        """Query the LIGO Channel Information System for the `Channel`
        matching the given name

        Parameters
        ----------
        name : `str`
            name of channel
        debug : `bool`, optional
            print verbose HTTP connection status for debugging,
            default: `False`
        timeout : `float`, optional
            maximum time to wait for a response from the CIS

        Returns
        -------
        Channel
             a new `Channel` containing all of the attributes set from
             its entry in the CIS
        """
        channellist = ChannelList.query(name, debug=debug, timeout=timeout)
        if len(channellist) == 0:
            raise ValueError("No channels found matching '%s'." % name)
        if len(channellist) > 1:
            raise ValueError("%d channels found matching '%s', please refine "
                             "search, or use `ChannelList.query` to return "
                             "all results." % (len(channellist), name))
        return channellist[0]

    @classmethod
    @with_import('nds2')
    def query_nds2(cls, name, host=None, port=None, connection=None,
                   type=None):
        """Query an NDS server for channel information

        Parameters
        ----------
        name : `str`
            name of requested channel
        host : `str`, optional
            name of NDS2 server.
        port : `int`, optional
            port number for NDS2 connection
        connection : `nds2.connection`
            open connection to use for query
        type : `str`, `int`
            NDS2 channel type with which to restrict query

        Returns
        -------
        channel : `Channel`
            channel with metadata retrieved from NDS2 server

        Raises
        ------
        ValueError
            if multiple channels are found for a given name

        Notes
        -----
        .. warning::

           A `host` is required if an open `connection` is not given
        """
        return ChannelList.query_nds2([name], host=host, port=port,
                                      connection=connection, type=type,
                                      unique=True)[0]

    @classmethod
    def from_nds2(cls, nds2channel):
        """Generate a new channel using an existing nds2.channel object
        """
        # extract metadata
        name = nds2channel.name
        sample_rate = nds2channel.sample_rate
        unit = nds2channel.signal_units
        if not unit:
            unit = None
        ctype = nds2channel.channel_type_to_string(nds2channel.channel_type)
        # get dtype
        dtype = {
            nds2channel.DATA_TYPE_INT16: numpy.int16,
            nds2channel.DATA_TYPE_INT32: numpy.int32,
            nds2channel.DATA_TYPE_INT64: numpy.int64,
            nds2channel.DATA_TYPE_FLOAT32: numpy.float32,
            nds2channel.DATA_TYPE_FLOAT64: numpy.float64,
            nds2channel.DATA_TYPE_COMPLEX32: numpy.complex64,
        }.get(nds2channel.data_type)
        return cls(name, sample_rate=sample_rate, unit=unit, dtype=dtype,
                   type=ctype)

    # -------------------------------------------------------------------------
    # methods

    @classmethod
    def parse_channel_name(cls, name, strict=True):
        """Decompose a channel name string into its components

        Parameters
        ----------
        name : `str`
            name to parse
        strict : `bool`, optional
            require exact matching of format, with no surrounding text,
            default `True`

        Returns
        -------
        match : `dict`
            `dict` of channel name components with the following keys:

            - `'ifo'`: the letter-number interferometer prefix
            - `'system'`: the top-level system name
            - `'subsystem'`: the second-level sub-system name
            - `'signal'`: the remaining underscore-delimited signal name
            - `'trend'`: the trend type
            - `'ndstype'`: the NDS2 channel suffix

            Any optional keys that aren't found will return a value of `None`

        Raises
        ------
        ValueError
            if the name cannot be parsed with at least an IFO and SYSTEM

        Examples
        --------
        >>> Channel.parse_channel_name('L1:LSC-DARM_IN1_DQ')
        {'ifo': 'L1',
         'ndstype': None,
         'signal': 'IN1_DQ',
         'subsystem': 'DARM',
         'system': 'LSC',
         'trend': None}

        >>> Channel.parse_channel_name(
            'H1:ISI-BS_ST1_SENSCOR_GND_STS_X_BLRMS_100M_300M.rms,m-trend')
        {'ifo': 'H1',
         'ndstype': 'm-trend',
         'signal': 'ST1_SENSCOR_GND_STS_X_BLRMS_100M_300M',
         'subsystem': 'BS',
         'system': 'ISI',
         'trend': 'rms'}
        """
        match = cls.MATCH.search(name)
        if match is None or (strict and (
                match.start() != 0 or match.end() != len(name))):
            raise ValueError("Cannot parse channel name according to LIGO "
                             "channel-naming convention T990033")
        return match.groupdict()

    def find_frametype(self, gpstime=None, frametype_match=None,
                       host=None, port=None, return_all=False,
                       exclude_tape=False):
        """Find the containing frametype(s) for this `Channel`

        Parameters
        ----------
        gpstime : `int`
            a reference GPS time at which to search for frame files
        frametype_match : `str`
            a regular expression string to use to down-select from the
            list of all available frametypes
        host : `str`
            the name of the datafind server to use for frame file discovery
        port : `int`
            the port of the datafind server on the given host
        return_all: `bool`, default: `False`
            return all matched frame types, otherwise only the first match is
            returned
        exclude_tape : `bool`, default: `False`
            do not search frame files that appear to be on magnetic tape

        Returns
        -------
        frametype : `str`, `list`
            the first matching frametype containing the this channel
            (`return_all=False`, or a `list` of all matches
        """
        return datafind.find_frametype(
            self, gpstime=gpstime, frametype_match=frametype_match,
            host=host, port=port, return_all=return_all,
            exclude_tape=exclude_tape)

    def copy(self):
        return type(self)(self.name, unit=self.unit,
                          sample_rate=self.sample_rate, dtype=self.dtype,
                          type=self.type, frametype=self.frametype,
                          model=self.model, url=self.url)

    def __str__(self):
        return self.name

    def __repr__(self):
        repr_ = '<Channel("%s"' % (str(self))
        if self.type:
            repr_ += ' [%s]' % self.type
        repr_ += ', %s' % self.sample_rate
        return repr_ + ') at %s>' % hex(id(self))

    def __eq__(self, other):
        for attr in ['name', 'sample_rate', 'unit', 'url', 'type', 'dtype']:
            try:
                if getattr(self, attr) != getattr(other, attr):
                    return False
            except TypeError:
                return False
        return True


_re_ifo = re.compile("[A-Z]\d:")
_re_cchar = re.compile("[-_]")


class ChannelList(list):
    """A `list` of `channels <Channel>`, with parsing utilities.
    """

    @property
    def ifos(self):
        """The `set` of interferometer prefixes used in this
        `ChannelList`.
        """
        return set([c.ifo for c in self])

    read = classmethod(reader(
        doc="""Read a `ChannelList` from a file

        Parameters
        ----------
        source : `str`, `file`
            either an open file object, or a file name path to read

        Notes
        -----"""))

    write = writer()

    @classmethod
    def from_names(cls, *names):
        new = cls()
        for namestr in names:
            for name in new._split_names(namestr):
                new.append(Channel(name))
        return new

    @staticmethod
    def _split_names(namestr):
        """Split a comma-separated list of channel names.
        """
        out = []
        namestr = re_quote.sub('', namestr)
        while True:
            namestr = namestr.strip('\' \n')
            if ',' not in namestr:
                break
            for nds2type in NDS2_CHANNEL_TYPE.keys() + ['']:
                if nds2type and ',%s' % nds2type in namestr:
                    try:
                        channel, ctype, namestr = namestr.split(',', 2)
                    except ValueError:
                        channel, ctype = namestr.split(',')
                        namestr = ''
                    out.append('%s,%s' % (channel, ctype))
                    break
                elif nds2type == '' and ',' in namestr:
                    channel, namestr = namestr.split(',', 1)
                    out.append(channel)
                    break
        if namestr:
            out.append(namestr)
        return out

    def find(self, name):
        """Find the `Channel` with a specific name in this `ChannelList`.

        Parameters
        ----------
        name : `str`
            name of the `Channel` to find

        Returns
        -------
        index : `int`
            the position of the first `Channel` in this `ChannelList`
            whose `~Channel.name` matches the search key.

        Raises
        ------
        ValueError
            if no matching `Channel` is found.
        """
        for i, chan in enumerate(self):
            if name == chan.name:
                return i
        raise ValueError(name)

    def sieve(self, name=None, sample_rate=None, sample_range=None,
              exact_match=False, **others):
        """Find all `Channels <Channel>` in this list matching the
        specified criteria.

        Parameters
        ----------
        name : `str`, or regular expression
            any part of the channel name against which to match
            (or full name if `exact_match=False` is given)
        sample_rate : `float`
            rate (number of samples per second) to match exactly
        sample_range : 2-`tuple`
            `[low, high]` closed interval or rates to match within
        exact_match : `bool`
            return channels matching `name` exactly, default: `False`

        Returns
        -------
        new : `ChannelList`
            a new `ChannelList` containing the matching channels
        """
        # format name regex
        if isinstance(name, re._pattern_type):
            flags = name.flags
            name = name.pattern
        else:
            flags = 0
        if exact_match:
            name = name.startswith(r'\A') and name or r"\A%s" % name
            name = name.endswith(r'\Z') and name or r"%s\Z" % name
        name_regexp = re.compile(name, flags=flags)
        c = list(self)
        if name is not None:
            c = [entry for entry in c if
                 name_regexp.search(entry.name) is not None]
        if sample_rate is not None:
            sample_rate = (isinstance(sample_rate, units.Quantity) and
                           sample_rate.value or float(sample_rate))
            c = [entry for entry in c if entry.sample_rate and
                 entry.sample_rate.value == sample_rate]
        if sample_range is not None:
            c = [entry for entry in c if
                 sample_range[0] <= float(entry.sample_rate) <=
                 sample_range[1]]
        for attr, val in others.iteritems():
            if val is not None:
                c = [entry for entry in c if
                     (hasattr(entry, attr) and getattr(entry, attr) == val)]

        return self.__class__(c)

    @classmethod
    def query(cls, name, debug=False, timeout=None):
        """Query the LIGO Channel Information System a `ChannelList`.

        Parameters
        ----------
        name : `str`
            name of channel, or part of it.
        debug : `bool`, optional
            print verbose HTTP connection status for debugging,
            default: `False`
        timeout : `float`, optional
            maximum time to wait for a response from the CIS

        Returns
        -------
        `ChannelList`
            a new list containing all `Channels <Channel>` found.
        """
        from .io import cis
        return cis.query(name, debug=debug, timeout=timeout)

    @classmethod
    @with_import('nds2')
    def query_nds2(cls, names, host=None, port=None, connection=None,
                   type=None, unique=False):
        """Query an NDS server for channel information

        Parameters
        ----------
        name : `str`
            name of requested channel
        host : `str`, optional
            name of NDS2 server.
        port : `int`, optional
            port number for NDS2 connection
        connection : `nds2.connection`
            open connection to use for query
        type : `str`, `int`
            NDS2 channel type with which to restrict query
        unique : `bool`, optional
            require a unique query result for each name given, default `False`

        Returns
        -------
        channellist : `ChannelList`
            list of `Channels <Channel>` with metadata retrieved from
            NDS2 server

        Raises
        ------
        ValueError
            if multiple channels are found for a given name and `unique=True`
            is given

        Notes
        -----
        .. warning::

           A `host` is required if an open `connection` is not given
        """
        from ..io.nds import (auth_connect, NDSWarning)
        out = cls()
        # connect
        if connection is None:
            if host is None:
                raise ValueError("Please given either an open nds2.connection,"
                                 " or the name of the host to connect to")
            connection = auth_connect(host, port)
        if isinstance(names, str):
            names = [names]
        for name in names:
            # format channel query
            if (type and (isinstance(type, (unicode, str)) or
                          (isinstance(type, int) and
                           log(type, 2).is_integer()))):
                channel = Channel(name, type=type)
            else:
                channel = Channel(name)
            if channel.ndstype is not None:
                found = connection.find_channels(
                    channel.ndsname, channel.ndstype)
            elif type is not None:
                found = connection.find_channels(channel.name, type)
            else:
                found = connection.find_channels(channel.name)
            found = ChannelList(map(Channel.from_nds2, found))
            _names = set([c.ndsname for c in found])
            if unique and len(_names) == 0:
                raise ValueError("No match for channel %r in NDS database"
                                 % name)
            if unique and len(_names) > 1:
                raise ValueError(
                    "Multiple matches for channel '%s' in NDS database, "
                    "ambiguous request:\n    %s"
                    % (name, '\n    '.join(['%s (%s, %s)'
                       % (str(c), c.type, c.sample_rate) for c in found])))
            elif unique and len(found) > 1:
                warnings.warn('Multiple instances of %r found with different '
                              'parameters, returning first.' % name,
                              NDSWarning)
                out.append(found[0])
            else:
                out.extend(found)
        return out

    @classmethod
    @with_import('nds2')
    def query_nds2_availability(cls, channels, start, end,
                                host, port=None):
        """Query for when data are available for these channels in NDS2

        Parameters
        ----------
        channels : `list`
            list of `Channel` or `str` for which to search
        start : `int`
            GPS start time of search, or any acceptable input to
            :meth:`~gwpy.time.to_gps`
        end : `int`
            GPS end time of search, or any acceptable input to
            :meth:`~gwpy.time.to_gps`
        host : `str`
            name of NDS server
        port : `int`, optional
            port number for NDS connection

        Returns
        -------
        nested dict
            a `dict` of (channel, `dict`) pairs where the nested `dict`
            is over (frametype, SegmentList) segment-availability pairs
        """
        from ..segments import (Segment, SegmentList, SegmentListDict)
        names = []
        for c in channels:
            name = str(c)
            if isinstance(c, Channel) and c.sample_rate:
                name += '%%%d' % c.sample_rate.value
            names.append(name)

        span = Segment(int(to_gps(start)), int(to_gps(end)))
        # build nds2_channel_source call
        cmd = ['nds2_channel_source', '-a',
               '-s', str(span[0]),
               '-n', host]
        if port is not None:
            cmd.extend(['-p', str(port)])
        cmd.extend(names)
        # call
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE)
        out, err = proc.communicate()
        retcode = proc.poll()
        if retcode:
            print(err, file=sys.stderr)
            raise subprocess.CalledProcessError(retcode, ' '.join(cmd),
                                                output=out)

        # parse output
        re_seg = re.compile('\A(?P<obs>[A-Z])-(?P<type>\w+):'
                            '(?P<start>\d+)-(?P<end>(\d+|inf))\Z')
        segs = {}
        i = 0
        for line in out.splitlines():
            if line.startswith('Requested source list:'):
                continue
            elif line.startswith('Error in daq'):
                raise RuntimeError(line)
            elif not line.startswith(re.split('[,%]', names[i])[0]):
                raise RuntimeError("Error parsing nds2_channel_source output")
            chan = channels[i]
            i += 1
            segs[chan] = SegmentListDict()
            for seg in line.split(' ', 1)[1].strip('{').rstrip('}').split(' '):
                try:
                    m = re_seg.search(seg).groupdict()
                except AttributeError:
                    if not seg:
                        continue
                    raise RuntimeError("Error parsing nds2_channel_source "
                                       "output:\n%s" % line)
                ftype = m['type']
                try:
                    seg = Segment(float(m['start']), float(m['end'])) & span
                except ValueError:
                    continue
                if ftype in segs[chan]:
                    segs[chan][ftype].append(seg)
                else:
                    segs[chan][ftype] = SegmentList([seg])
            segs[chan].coalesce()
        return segs
